"""
PublicConfigurationSerializer
"""
from waves_gateway.common import Injectable
from waves_gateway.model import PublicConfiguration


@Injectable()
class PublicConfigurationSerializer(object):
    """
    Defines the conversion of a PublicConfiguration instance into a dict object.
    The resulting object may be delivered to the Web Application.
    """

    def as_dict(self, value: PublicConfiguration) -> dict:
        res = dict()

        res[PublicConfiguration.CUSTOM_CURRENCY_NAME_DICT_KEY] = value.custom_currency_name
        res[PublicConfiguration.GATEWAY_WAVES_ADDRESS_DICT_KEY] = value.gateway_waves_address
        res[PublicConfiguration.GATEWAY_COIN_HOLDER_DICT_KEY] = value.gateway_coin_holder
        res[PublicConfiguration.WAVES_NODE_DICT_KEY] = value.waves_node
        res[PublicConfiguration.WAVES_ASSET_ID] = value.waves_asset_id
        res[PublicConfiguration.WEB_PRIMARY_COLOR] = value.web_primary_color

        if value.coin_transaction_web_link is not None:
            res[PublicConfiguration.COIN_TRANSACTION_WEB_LINK] = value.coin_transaction_web_link

        if value.coin_address_web_link is not None:
            res[PublicConfiguration.COIN_ADDRESS_WEB_LINK] = value.coin_address_web_link

        res[PublicConfiguration.WAVES_ADDRESS_WEB_LINK] = value.waves_address_web_link
        res[PublicConfiguration.WAVES_TRANSACTION_WEB_LINK] = value.waves_transaction_web_link

        return res



"""
TransactionAttemptListSerializer
"""

from typing import Any, Union

from waves_gateway.model import TransactionAttemptList, TransactionAttempt, TransactionAttemptReceiver, \
    AttemptListTrigger, TransactionSender

from waves_gateway.common import map_array, Injectable


@Injectable()
class TransactionAttemptListSerializer(object):
    """
    Defines a conversion from an TransactionAttemptList instance into a dict instance
    and backwards. The dict instance can be used for persistence or an external API.
    """

    # noinspection PyMethodMayBeStatic
    def _receiver_as_dict(self, receiver: TransactionAttemptReceiver) -> dict:
        res = dict()  # type: Any

        res[TransactionAttemptReceiver.DICT_ADDRESS_KEY] = receiver.address
        res[TransactionAttemptReceiver.DICT_AMOUNT_KEY] = str(receiver.amount)

        return res

    def _attempt_as_dict(self, attempt: TransactionAttempt) -> dict:
        res = dict()  # type: Any

        res[TransactionAttempt.DICT_CURRENCY] = attempt.currency
        res[TransactionAttempt.DICT_FEE] = str(attempt.fee)
        res[TransactionAttempt.DICT_SENDER] = attempt.sender
        res[TransactionAttempt.DICT_RECEIVERS] = map_array(self._receiver_as_dict, attempt.receivers)

        return res

    # noinspection PyMethodMayBeStatic
    def _trigger_as_dict(self, trigger: AttemptListTrigger):
        res = dict()

        res[AttemptListTrigger.DICT_RECEIVER] = trigger.receiver
        res[AttemptListTrigger.DICT_TX] = trigger.tx
        res[AttemptListTrigger.DICT_CURRENCY] = trigger.currency

        if trigger.senders is not None:
            res[AttemptListTrigger.DICT_SENDERS_KEY] = map_array(self._sender_as_dict, trigger.senders)

        return res

    def attempt_list_as_dict(self, attempt_list: TransactionAttemptList):
        res = dict()

        res[TransactionAttemptList.DICT_ID] = attempt_list.attempt_list_id
        res[TransactionAttemptList.DICT_TRANSACTIONS] = attempt_list.transactions
        res[TransactionAttemptList.DICT_ATTEMPTS] = map_array(self._attempt_as_dict, attempt_list.attempts)
        res[TransactionAttemptList.DICT_TRIGGER] = self._trigger_as_dict(attempt_list.trigger)
        res[TransactionAttemptList.DICT_TRIES] = attempt_list.tries

        if attempt_list.last_modified is not None:
            res[TransactionAttemptList.DICT_LAST_MODIFIED] = attempt_list.last_modified

        if attempt_list.created_at is not None:
            res[TransactionAttemptList.DICT_CREATED_AT] = attempt_list.created_at

        return res

    # noinspection PyMethodMayBeStatic
    def _receiver_from_dict(self, data: dict) -> TransactionAttemptReceiver:
        return TransactionAttemptReceiver(
            amount=int(data[TransactionAttemptReceiver.DICT_AMOUNT_KEY]),
            address=data[TransactionAttemptReceiver.DICT_ADDRESS_KEY])

    # noinspection PyMethodMayBeStatic
    def _sender_from_dict(self, data: dict) -> TransactionSender:
        return TransactionSender(address=data[TransactionSender.DICT_ADDRESS_KEY])

    def _sender_as_dict(self, sender: TransactionSender) -> dict:
        res = dict()
        res[TransactionSender.DICT_ADDRESS_KEY] = sender.address
        return res

    def _attempt_from_dict(self, data: dict) -> TransactionAttempt:
        return TransactionAttempt(
            currency=data[TransactionAttempt.DICT_CURRENCY],
            fee=int(data[TransactionAttempt.DICT_FEE]),
            sender=data[TransactionAttempt.DICT_SENDER],
            receivers=map_array(self._receiver_from_dict, data[TransactionAttempt.DICT_RECEIVERS]))

    # noinspection PyMethodMayBeStatic
    def _trigger_from_dict(self, data: dict) -> AttemptListTrigger:
        if AttemptListTrigger.DICT_SENDERS_KEY in data:
            senders = map_array(self._sender_from_dict, data[AttemptListTrigger.DICT_SENDERS_KEY])
        else:
            senders = None

        return AttemptListTrigger(
            tx=data[AttemptListTrigger.DICT_TX],
            receiver=data[AttemptListTrigger.DICT_RECEIVER],
            currency=data[AttemptListTrigger.DICT_CURRENCY],
            senders=senders)

    def attempt_list_from_dict(self, data: dict) -> TransactionAttemptList:
        created_at = None
        last_modified = None
        tries = TransactionAttemptList.DEFAULT_TRIES

        if TransactionAttemptList.DICT_CREATED_AT in data:
            created_at = data[TransactionAttemptList.DICT_CREATED_AT]

        if TransactionAttemptList.DICT_LAST_MODIFIED in data:
            last_modified = data[TransactionAttemptList.DICT_LAST_MODIFIED]

        if TransactionAttemptList.DICT_TRIES in data:
            tries = data[TransactionAttemptList.DICT_TRIES]

        return TransactionAttemptList(
            trigger=self._trigger_from_dict(data[TransactionAttemptList.DICT_TRIGGER]),
            attempts=map_array(self._attempt_from_dict, data[TransactionAttemptList.DICT_ATTEMPTS]),
            transactions=data[TransactionAttemptList.DICT_TRANSACTIONS],
            last_modified=last_modified,
            created_at=created_at,
            tries=tries,
            attempt_list_id=data[TransactionAttemptList.DICT_ID])



"""
GatewayLoggingConfigurationService
"""

from logging import Logger, Handler
from typing import Optional, Union, List

from waves_gateway.common import Injectable, LOGGING_HANDLER_LIST, MANAGED_LOGGER_LIST


@Injectable(deps=[Logger, LOGGING_HANDLER_LIST, MANAGED_LOGGER_LIST])
class GatewayLoggingConfigurationService(object):
    """
    Controls the logging of the Gateway.
    """

    def __init__(self, logger: Logger, handlers: List[Handler], managed_loggers: List[Logger]) -> None:
        self._logger = logger
        self._logging_handlers = handlers
        self._managed_loggers = managed_loggers

    def set_log_level(self, level, loggers: Optional[Union[List[str], str]] = None):
        """
        Adjusts the log level for all known loggers or specific ones if an additional parameter is added.
        """
        to_be_adjusted = list()  # type: List[Logger]

        if loggers is None:
            for logger in self._managed_loggers:
                to_be_adjusted.append(logger)

            to_be_adjusted.append(self._logger)
        elif isinstance(loggers, str):
            if loggers == self._logger.name:
                to_be_adjusted.append(self._logger)

            for logger in self._managed_loggers:
                if logger.name == loggers:
                    to_be_adjusted.append(logger)
        else:
            if self._logger.name in loggers:
                to_be_adjusted.append(self._logger)

            for logger in self._managed_loggers:
                if logger.name in loggers:
                    to_be_adjusted.append(logger)

        for logger in to_be_adjusted:
            logger.setLevel(level)

        for handler in self._logging_handlers:
            handler.setLevel(level)



"""
PublicConfiguration
"""

from typing import Optional


class PublicConfiguration:
    """
    Part of the configuration that is allowed to be available to the public,
    for example via ReST.
    """

    CUSTOM_CURRENCY_NAME_DICT_KEY = 'custom_currency_name'
    GATEWAY_WAVES_ADDRESS_DICT_KEY = 'gateway_waves_address'
    GATEWAY_COIN_HOLDER_DICT_KEY = 'gateway_coin_holder'
    WAVES_NODE_DICT_KEY = 'waves_node'
    WAVES_ASSET_ID = 'waves_asset_id'
    COIN_TRANSACTION_WEB_LINK = 'coin_transaction_web_link'
    WAVES_TRANSACTION_WEB_LINK = 'waves_transaction_web_link'
    COIN_ADDRESS_WEB_LINK = 'coin_address_web_link'
    WAVES_ADDRESS_WEB_LINK = 'waves_address_web_link'
    WEB_PRIMARY_COLOR = 'web_primary_color'

    def __init__(self,
                 custom_currency_name: str,
                 gateway_waves_address: str,
                 gateway_coin_address: str,
                 waves_node: str,
                 waves_asset_id: str,
                 waves_transaction_web_link: str,
                 waves_address_web_link: str,
                 coin_transaction_web_link: Optional[str] = None,
                 coin_address_web_link: Optional[str] = None,
                 web_primary_color: Optional[str] = None) -> None:
        self._custom_currency_name = custom_currency_name
        self._gateway_waves_address = gateway_waves_address
        self._gateway_coin_holder = gateway_coin_address
        self._waves_node = waves_node
        self._waves_asset_id = waves_asset_id
        self._coin_transaction_web_link = coin_transaction_web_link
        self._waves_transaction_web_link = waves_transaction_web_link
        self._waves_address_web_link = waves_address_web_link
        self._coin_address_web_link = coin_address_web_link
        self._web_primary_color = web_primary_color

    @property
    def custom_currency_name(self) -> str:
        return self._custom_currency_name

    @property
    def gateway_waves_address(self) -> str:
        return self._gateway_waves_address

    @property
    def gateway_coin_holder(self) -> str:
        return self._gateway_coin_holder

    @property
    def waves_node(self) -> str:
        return self._waves_node

    @property
    def waves_asset_id(self) -> str:
        return self._waves_asset_id

    @property
    def waves_transaction_web_link(self) -> str:
        return self._waves_transaction_web_link

    @property
    def coin_address_web_link(self) -> Optional[str]:
        return self._coin_address_web_link

    @property
    def coin_transaction_web_link(self) -> Optional[str]:
        return self._coin_transaction_web_link

    @property
    def waves_address_web_link(self) -> str:
        return self._waves_address_web_link

    @property
    def web_primary_color(self) -> str:
        return self._web_primary_color



"""
Contains abstractions for all data structures that this Gateway handles.
"""

from .transaction import Transaction, TransactionReceiver, TransactionSender
from .mapping_entry import MappingEntry
from .key_pair import KeyPair
from .public_configuration import PublicConfiguration
from .transaction_attempt import TransactionAttempt, TransactionAttemptReceiver
from .transaction_attempt_list import TransactionAttemptList, AttemptListTrigger
from .attempt_list_query import AttemptListQuery
from .polling_delay_config import PollingDelayConfig
from .gateway_config_file import GatewayConfigFile
from .polling_state import PollingState, PollingTransactionState



"""
Defines storage interfaces and default implementations.
"""

from .map_storage import MapStorage
from .wallet_storage import WalletStorage
from .mongo_transaction_attempt_list_storage_impl import MongoTransactionAttemptListStorageImpl
from .mongo_map_storage_impl import MongoMapStorageImpl
from .mongo_wallet_storage_impl import MongoWalletStorageImpl
from .mongo_key_value_storage_impl import MongoKeyValueStorageImpl
from .coin_block_height_storage_proxy_impl import CoinBlockHeightStorageProxyImpl
from .waves_block_height_storage_proxy_impl import WavesBlockHeightStorageProxyImpl
from .key_value_storage import KeyValueStorage
from .block_height_storage_proxy import BlockHeightStorageProxy
from .transaction_attempt_list_storage import TransactionAttemptListStorage
from .coin_polling_state_storage_proxy_impl import CoinPollingStateStorageProxyImpl
from .waves_polling_state_storage_proxy_impl import WavesPollingStateStorageProxyImpl
from .polling_state_storage_proxy import PollingStateStorageProxy



"""
GatewayController
"""

from abc import ABC, abstractmethod
from typing import Optional, List

from waves_gateway.model import AttemptListTrigger, TransactionAttemptList, AttemptListQuery


class GatewayController(ABC):
    """
    Defines the API for the Waves Client application.
    All possible interfaces should forward their requests to this abstract controller.
    """

    @abstractmethod
    def create_address(self, waves_address: str) -> str:
        """
        Returns a new address of the custom currency or an existing one
        if the waves_address is already associated to an address of the custom
        cryptocurrency.
        """
        pass

    @abstractmethod
    def get_attempt_list_by_trigger(self, trigger: AttemptListTrigger) -> Optional[TransactionAttemptList]:
        """
        Find an attempt_list by its trigger. The result may be None if no such trigger exists.
        """
        pass

    @abstractmethod
    def get_attempt_list_by_id(self, attempt_list_id: str) -> Optional[TransactionAttemptList]:
        """
        Find an attempt_list by its id. The result may be None if no such id exists.
        """
        pass

    @abstractmethod
    def query_attempt_lists(self, query: AttemptListQuery) -> List[TransactionAttemptList]:
        """
        Query attempt lists by specific criteria.
        """
        pass

    @abstractmethod
    def validate_waves_address(self, address: str) -> bool:
        """
        Tests if the given address is an valid waves address.
        """
        pass

    @abstractmethod
    def check_waves_transaction(self, tx: str) -> None:
        """
        Checks the given waves transaction whether it requires a new attempt list to be created.
        """
        pass

    @abstractmethod
    def check_coin_transaction(self, tx: str) -> None:
        """
        Checks the given coin transaction whether it requires a new attempt list to be created.
        """
        pass



"""GatewayConfigParser"""

from configparser import ConfigParser
from decimal import Decimal
from typing import Union, Optional
from waves_gateway.common import Injectable, KeyPair, InvalidConfigError
from waves_gateway.model import GatewayConfigFile


@Injectable()
class GatewayConfigParser(object):
    """
    May be used to parse a gateway configuration file.
    """

    def _parse_node_section(self, config_parser: ConfigParser, parsed_config: GatewayConfigFile) -> None:
        parsed_config.waves_node = config_parser.get('node', 'waves')
        parsed_config.coin_node = config_parser.get('node', 'coin')

    def _parse_number(self, config: ConfigParser, section: str, option: str) -> Union[int, Decimal]:
        """Parses a number field to an int or Decimal."""
        value = config.get(section, option)  # type: str
        if '.' in value:
            return Decimal(value)
        else:
            return int(value)

    def _parse_fee_section(self, config_parser: ConfigParser, parsed_config: GatewayConfigFile) -> None:
        parsed_config.coin_fee = self._parse_number(config_parser, 'fee', 'coin')
        parsed_config.gateway_fee = self._parse_number(config_parser, 'fee', 'gateway')

    def _parse_gateway_address_section(self, config_parser: ConfigParser, parsed_config: GatewayConfigFile):
        parsed_config.gateway_owner_address = config_parser.get('gateway_address', 'owner')

        gateway_waves_address_public_key = config_parser.get('gateway_address', 'waves_public_key')
        gateway_waves_address_private_key = config_parser.get('gateway_address', 'waves_private_key')
        gateway_coin_address_public_key = config_parser.get('gateway_address', 'coin_public_key')
        gateway_coin_address_private_key = config_parser.get(
            'gateway_address', 'coin_private_key', fallback=None)  # type: Optional[str]

        parsed_config.gateway_waves_address_secret = KeyPair(
            public=gateway_waves_address_public_key, secret=gateway_waves_address_private_key)
        parsed_config.gateway_coin_address_secret = KeyPair(
            public=gateway_coin_address_public_key, secret=gateway_coin_address_private_key)

    def _parse_mongodb_section(self, config_parser: ConfigParser, parsed_config: GatewayConfigFile) -> None:
        parsed_config.mongo_host = config_parser.get('mongodb', 'host')
        parsed_config.mongo_port = config_parser.getint('mongodb', 'port')
        parsed_config.mongo_database = config_parser.get('mongodb', 'database')

    def _parse_other_section(self, config_parser: ConfigParser, parsed_config: GatewayConfigFile) -> None:
        parsed_config.waves_chain = config_parser.get('other', 'waves_chain', fallback="mainnet")
        parsed_config.environment = config_parser.get('other', 'environment', fallback="prod")

    def _parse_server_section(self, config_parser: ConfigParser, parsed_config: GatewayConfigFile):
        parsed_config.gateway_host = config_parser.get('server', 'host', fallback='localhost')
        parsed_config.gateway_port = config_parser.getint('server', 'port', fallback=5000)

    def parse_config_file_content(self, file_content: str) -> GatewayConfigFile:
        """Parses the given config file."""
        config_parser = ConfigParser()
        config_parser.read_string(file_content)
        parsed_config = GatewayConfigFile()

        self._assert_not_old_format(config_parser)

        self._parse_node_section(config_parser, parsed_config)

        if config_parser.has_section('fee'):
            self._parse_fee_section(config_parser, parsed_config)

        self._parse_gateway_address_section(config_parser, parsed_config)

        if config_parser.has_section('mongodb'):
            self._parse_mongodb_section(config_parser, parsed_config)

        if config_parser.has_section('other'):
            self._parse_other_section(config_parser, parsed_config)

        self._parse_server_section(config_parser, parsed_config)

        return parsed_config

    def _assert_not_old_format(self, config_parser: ConfigParser):
        """Asserts that the old format in the ltc gateway is no longer used."""

        def assert_option_does_not_exist(section: str, option: str):
            """Raises an exception if the selected option exists."""
            if config_parser.has_section(section) and config_parser.has_option(section, option):
                raise InvalidConfigError('Detected use of option ' + option + ' in the section ' + section +
                                         '. Using ltc in a configuration file is deprecated. Please use coin instead!')

        assert_option_does_not_exist('node', 'ltc')
        assert_option_does_not_exist('fee', 'ltc')
        assert_option_does_not_exist('gateway_address', 'ltc_public_key')
        assert_option_does_not_exist('gateway_address', 'ltc_private_key')

    def parse_config_file(self, file_path: str):
        """First reads the config file and then calls parse_config_file_content with the content."""
        file = open(file_path, "r")
        return self.parse_config_file_content(file.read())



"""
Gateway
"""

import logging
import os
from typing import Optional, List

import flask  # type: ignore
import pywaves  # type: ignore

from flask import Flask

from waves_gateway.service import COIN_TRANSACTION_POLLING_SERVICE, \
    WAVES_TRANSACTION_POLLING_SERVICE, WavesTransactionConsumerImpl, CoinTransactionConsumerImpl, IntegerConverterService
from waves_gateway.common import Factory, LOGGING_HANDLER_LIST, MANAGED_LOGGER_LIST, POLLING_DELAY_SECONDS, \
    CUSTOM_CURRENCY_NAME, GATEWAY_OWNER_ADDRESS, WALLET_STORAGE_COLLECTION_NAME, MAP_STORAGE_COLLECTION_NAME, \
    KEY_VALUE_STORAGE_COLLECTION_NAME, TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION_NAME, GATEWAY_COIN_ADDRESS_SECRET, \
    GATEWAY_COIN_ADDRESS, WAVES_NODE, WAVES_ASSET_ID, WAVES_CHAIN, ONLY_ONE_TRANSACTION_RECEIVER, \
    COIN_TRANSACTION_WEB_LINK, WAVES_TRANSACTION_WEB_LINK, COIN_ADDRESS_WEB_LINK, WAVES_ADDRESS_WEB_LINK, \
    ATTEMPT_LIST_MAX_COMPLETION_TRIES, GATEWAY_WAVES_ADDRESS_SECRET, GATEWAY_WAVES_ADDRESS, NUM_ATTEMPT_LIST_WORKERS, \
    GATEWAY_HOST, GATEWAY_PORT, WALLET_STORAGE_COLLECTION, MAP_STORAGE_COLLECTION, KEY_VALUE_STORAGE_COLLECTION, \
    TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION, GATEWAY_PYWAVES_ADDRESS, COIN_CHAIN_QUERY_SERVICE_CONVERTER_PROXY, \
    WAVES_CHAIN_QUERY_SERVICE_CONVERTER_PROXY, FLASK_NAME, COIN_MAX_HANDLE_TRANSACTION_TRIES, \
    WAVES_MAX_HANDLE_TRANSACTION_TRIES, WAVES_LAST_BLOCK_DISTANCE, COIN_LAST_BLOCK_DISTANCE, WEB_PRIMARY_COLOR
from waves_gateway.model import PollingDelayConfig
from waves_gateway.factory import CoinAddressFactory

import waves_gateway.common as common
import waves_gateway.controller as controller
import waves_gateway.model as model
import waves_gateway.service as service
import waves_gateway.storage as storage
import waves_gateway.factory as factory
from pymongo.database import Database as MongoDatabase  # type: ignore

from waves_gateway.model import PublicConfiguration
from waves_gateway.service import COIN_CHAIN_QUERY_SERVICE, COIN_ADDRESS_VALIDATION_SERVICE, COIN_TRANSACTION_SERVICE, COIN_INTEGER_CONVERTER_SERVICE, \
    ASSET_INTEGER_CONVERTER_SERVICE
from waves_gateway.storage import CoinBlockHeightStorageProxyImpl, WavesBlockHeightStorageProxyImpl, CoinPollingStateStorageProxyImpl, WavesPollingStateStorageProxyImpl, PollingStateStorageProxy


@Factory(Flask, deps=[FLASK_NAME])
def flask_factory(flask_name: str):
    """Configures the flask instance."""
    directory, file = os.path.split(os.path.realpath(__file__))
    return flask.Flask(flask_name, static_folder=os.path.join(directory, 'static'))


@Factory(logging.Logger, deps=[Flask, LOGGING_HANDLER_LIST, MANAGED_LOGGER_LIST])
def logger_factory(flask: Flask, logging_handlers: List[logging.Handler], managed_loggers: List[logging.Logger]):
    """Creates the major logger instance of the Gateway."""
    logger = flask.logger
    logger.setLevel(logging.DEBUG)

    for handler in logging_handlers:
        for managed_logger in managed_loggers:
            managed_logger.addHandler(handler)

        logger.addHandler(handler)

    return logger


@Factory(LOGGING_HANDLER_LIST, weak=True)
def logging_handlers_factory():
    """Provides a default list of logging handlers that log to the console."""
    logging_handlers = list()
    formatter = logging.Formatter("[%(asctime)s] %(levelname)s {%(name)s} - %(message)s")

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG)
    stream_handler.setFormatter(formatter)

    logging_handlers.append(stream_handler)
    return logging_handlers


@Factory(PollingDelayConfig, opt_deps=[POLLING_DELAY_SECONDS], weak=True)
def polling_delay_config_factory(polling_delay_s: Optional[float]):
    """
    Provides a default PollingDelayConfig.
    If the old API (polling_delay_s) is used, the PollingDelayConfig is constructed using the polling_delay_s value.
    """
    if polling_delay_s is None:
        return model.PollingDelayConfig()
    elif polling_delay_s is not None:
        return model.PollingDelayConfig.from_single_polling_delay(polling_delay_s)
    else:
        raise common.InvalidConfigError(
            'It is not possible to specify both a fixed polling_delay_s and a dynamic PollingDelayConfig.')


@Factory(
    TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION, deps=[TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION_NAME, MongoDatabase])
@Factory(WALLET_STORAGE_COLLECTION, deps=[WALLET_STORAGE_COLLECTION_NAME, MongoDatabase])
@Factory(MAP_STORAGE_COLLECTION, deps=[MAP_STORAGE_COLLECTION_NAME, MongoDatabase])
@Factory(KEY_VALUE_STORAGE_COLLECTION, deps=[KEY_VALUE_STORAGE_COLLECTION_NAME, MongoDatabase])
def collection_factory(collection_name: str, database: MongoDatabase):
    """Creates a new collection with the given name."""
    return database.get_collection(collection_name)


@Factory(GATEWAY_PYWAVES_ADDRESS, deps=[GATEWAY_WAVES_ADDRESS_SECRET, WAVES_NODE, WAVES_CHAIN])
def gateway_pywaves_address(gateway_waves_address_secret: model.KeyPair, waves_node: str, waves_chain: str):
    """Creates an address instance from the pywaves library that represents the Gateway waves address."""
    pywaves.setNode(waves_node, waves_chain)
    return pywaves.Address(gateway_waves_address_secret.public, privateKey=gateway_waves_address_secret.secret)


@Factory(
    COIN_TRANSACTION_POLLING_SERVICE,
    deps=[
        COIN_CHAIN_QUERY_SERVICE_CONVERTER_PROXY, logging.Logger, CoinBlockHeightStorageProxyImpl,
        CoinTransactionConsumerImpl, model.PollingDelayConfig, CoinPollingStateStorageProxyImpl,
        COIN_MAX_HANDLE_TRANSACTION_TRIES, COIN_LAST_BLOCK_DISTANCE
    ])
def coin_transaction_polling_service_factory(coin_chain_query_service_converter_proxy, logger,
                                             coin_block_height_storage_proxy, coin_transaction_consumer,
                                             polling_delay_config, polling_state_storage: PollingStateStorageProxy,
                                             max_handle_transaction_tries: int, last_block_distance):
    """Creates a TransactionPollingService instance with the necessary dependencies for the custom cryptocurrency."""
    return service.TransactionPollingService(
        chain_query_service=coin_chain_query_service_converter_proxy,
        logger=logger,
        block_height_storage=coin_block_height_storage_proxy,
        transaction_consumer=coin_transaction_consumer,
        max_polling_delay_s=polling_delay_config.coin_max_polling_delay_s,
        min_polling_delay_s=polling_delay_config.coin_min_polling_delay_s,
        max_handle_transaction_tries=max_handle_transaction_tries,
        polling_state_storage=polling_state_storage,
        last_block_distance=last_block_distance)


@Factory(
    WAVES_TRANSACTION_POLLING_SERVICE,
    deps=[
        WAVES_CHAIN_QUERY_SERVICE_CONVERTER_PROXY, logging.Logger, WavesBlockHeightStorageProxyImpl,
        WavesTransactionConsumerImpl, model.PollingDelayConfig, WavesPollingStateStorageProxyImpl,
        WAVES_MAX_HANDLE_TRANSACTION_TRIES, WAVES_LAST_BLOCK_DISTANCE
    ])
def waves_transaction_polling_service_factory(waves_chain_query_service_converter_proxy, logger,
                                              waves_block_height_storage_proxy, waves_transaction_consumer,
                                              polling_delay_config, polling_state_storage: PollingStateStorageProxy,
                                              max_handle_transaction_tries: int, last_block_distance: int):
    """Creates a TransactionPollingService instance with the necessary dependencies for Waves."""
    return service.TransactionPollingService(
        chain_query_service=waves_chain_query_service_converter_proxy,
        logger=logger,
        block_height_storage=waves_block_height_storage_proxy,
        transaction_consumer=waves_transaction_consumer,
        max_polling_delay_s=polling_delay_config.waves_max_polling_delay_s,
        min_polling_delay_s=polling_delay_config.waves_min_polling_delay_s,
        max_handle_transaction_tries=max_handle_transaction_tries,
        polling_state_storage=polling_state_storage,
        last_block_distance=last_block_distance)


@Factory(
    PublicConfiguration,
    deps=[
        CUSTOM_CURRENCY_NAME, GATEWAY_WAVES_ADDRESS, GATEWAY_COIN_ADDRESS, WAVES_NODE, WAVES_ASSET_ID,
        WAVES_TRANSACTION_WEB_LINK, WAVES_ADDRESS_WEB_LINK, WEB_PRIMARY_COLOR
    ],
    opt_deps=[COIN_TRANSACTION_WEB_LINK, COIN_ADDRESS_WEB_LINK])
def public_configuration_factory(custom_currency_name: str,
                                 gateway_waves_address: str,
                                 gateway_coin_address: str,
                                 waves_node: str,
                                 waves_asset_id: str,
                                 waves_transaction_web_link: str,
                                 waves_address_web_link: str,
                                 web_primary_color: str,
                                 coin_transaction_web_link: Optional[str] = None,
                                 coin_address_web_link: Optional[str] = None):
    """Creates a constant global configuration instance that may be provided to the client."""
    return model.PublicConfiguration(
        custom_currency_name=custom_currency_name,
        gateway_waves_address=gateway_waves_address,
        gateway_coin_address=gateway_coin_address,
        waves_node=waves_node,
        waves_asset_id=waves_asset_id,
        waves_transaction_web_link=waves_transaction_web_link,
        waves_address_web_link=waves_address_web_link,
        coin_transaction_web_link=coin_transaction_web_link,
        coin_address_web_link=coin_address_web_link,
        web_primary_color=web_primary_color)


@Factory(COIN_INTEGER_CONVERTER_SERVICE, weak=True)
def coin_integer_converter_service_factory():
    """Provides the default IntegerConverterService for the custom cryptocurrency."""
    return IntegerConverterService()


@Factory(ASSET_INTEGER_CONVERTER_SERVICE, weak=True)
def asset_integer_converter_service_factory():
    """Provides the default IntegerConverterService for the asset."""
    return IntegerConverterService()


@Factory(WAVES_ADDRESS_WEB_LINK, weak=True, deps=[WAVES_CHAIN])
def waves_address_web_link_factory(waves_chain: str):
    """Provides default value for WAVES_ADDRESS_WEB_LINK depending on WAVES_CHAIN."""
    if waves_chain == 'testnet':
        return 'https://testnet.wavesexplorer.com/address/{{address}}'
    else:
        return 'https://wavesexplorer.com/address/{{address}}'


@Factory(WAVES_TRANSACTION_WEB_LINK, weak=True, deps=[WAVES_CHAIN])
def waves_transaction_web_link_factory(waves_chain: str):
    """Provides default value for WAVES_TRANSACTION_WEB_LINK depending on WAVES_CHAIN."""
    if waves_chain == 'testnet':
        return 'https://testnet.wavesexplorer.com/tx/{{tx}}'
    else:
        return 'https://wavesexplorer.com/tx/{{tx}}'


class Gateway(object):
    """
    Provides the configuration for the Gateway.
    When setting up the Gateway for a specific currency, use this class
    to collect the information and call the run method of the instance.

    The constructor has specific parameters that are marked as required.
    Those parameters must be provided for a minimum configuration.
    To provide those instances, the abstract classes have to be implemented.
    Then, create one instance per each and overhand them to the gateway.
    Provided instances should not have any dependencies on other instances managed by this class.
    But, if strongly necessary they may use the public instances provided by an instance of the Gateway class.

    Alongside the required parameters, there are additional optional parameters that may be provided.
    The Gateway is capable of filling them with default implementations listed in the constructor documentation.
    But, if an default implementation is not suitable, it may be replaced by handing over a different implementation
    to the Gateway. If so, this implementation will be used and no instance of an default implementation will
    be created.

    The resulting Gateway instance is not running. Which means that the pollers are not started.
    To start those pollers call the run method on the Gateway instance.
    Furthermore, it is possible to stop the pollers by calling cancel on the Gateway. But, it should generally
    not be necessary to do so.
    """

    DEFAULT_FLASK_NAME = 'waves-gw'
    DEFAULT_CUSTOM_CURRENCY_NAME = 'Custom Currency'
    DEFAULT_POLLING_DELAY = 0
    DEFAULT_WAVES_POLLING_DELAY = 0
    DEFAULT_HOSTNAME = 'localhost'
    DEFAULT_PORT = 5000
    DEFAULT_WAVES_CHAIN = 'mainnet'
    DEFAULT_NUM_ATTEMPT_LIST_WORKERS = 1
    DEFAULT_MAX_COMPLETION_TRIES = 30
    DEFAULT_WALLET_STORAGE_COLLECTION_NAME = 'wallet'
    DEFAULT_MAP_STORAGE_COLLECTION_NAME = 'mapping'
    DEFAULT_ATTEMPT_LIST_STORAGE_COLLECTION_NAME = 'attempt_list'
    DEFAULT_KEY_VALUE_STORAGE_COLLECTION_NAME = 'key_value'
    DEFAULT_COIN_MAX_HANDLE_TRANSACTION_TRIES = 5
    DEFAULT_WAVES_MAX_HANDLE_TRANSACTION_TRIES = 5
    DEFAULT_COIN_LAST_BLOCK_DISTANCE = 0
    DEFAULT_WAVES_LAST_BLOCK_DISTANCE = 1
    DEFAULT_WEB_PRIMARY_COLOR = '#2196f3'

    def _init_managed_loggers(self, managed_loggers: List[str]) -> None:
        res = list()  # type: List[logging.Logger]

        if managed_loggers is None:
            managed_loggers = list()

        if "werkzeug" not in managed_loggers:
            managed_loggers.append("werkzeug")

        for logger_name in managed_loggers:
            logger_instance = logging.getLogger(logger_name)
            logger_instance.handlers = []
            logger_instance.setLevel(logging.DEBUG)
            res.append(logger_instance)

        self._injector.overwrite(MANAGED_LOGGER_LIST, res)

    def __init__(self,
                 coin_address_factory: factory.CoinAddressFactory,
                 coin_chain_query_service: service.ChainQueryService,
                 gateway_waves_address_secret: model.KeyPair,
                 coin_transaction_service: service.TransactionService,
                 gateway_coin_address_secret: model.KeyPair,
                 waves_node: str,
                 waves_asset_id: str,
                 gateway_owner_address: str,
                 fee_service: service.FeeService,
                 coin_address_validation_service: service.AddressValidationService,
                 coin_transaction_web_link: Optional[str] = None,
                 waves_transaction_web_link: Optional[str] = None,
                 coin_address_web_link: Optional[str] = None,
                 waves_address_web_link: Optional[str] = None,
                 coin_integer_converter_service: Optional[service.IntegerConverterService] = None,
                 asset_integer_converter_service: Optional[service.IntegerConverterService] = None,
                 waves_chain=DEFAULT_WAVES_CHAIN,
                 only_one_transaction_receiver: bool = False,
                 custom_currency_name: str = DEFAULT_CUSTOM_CURRENCY_NAME,
                 logging_handlers: Optional[list] = None,
                 managed_loggers: List[str] = list(),
                 attempt_list_max_completion_tries: int = DEFAULT_MAX_COMPLETION_TRIES,
                 host: str = DEFAULT_HOSTNAME,
                 port: int = DEFAULT_PORT,
                 num_attempt_list_workers=DEFAULT_NUM_ATTEMPT_LIST_WORKERS,
                 polling_delay_s: Optional[float] = None,
                 polling_delay_config: Optional[model.PollingDelayConfig] = None,
                 transaction_attempt_list_storage_collection_name=DEFAULT_ATTEMPT_LIST_STORAGE_COLLECTION_NAME,
                 wallet_storage_collection_name: str = DEFAULT_WALLET_STORAGE_COLLECTION_NAME,
                 map_storage_collection_name: str = DEFAULT_MAP_STORAGE_COLLECTION_NAME,
                 key_value_storage_collection_name: str = DEFAULT_KEY_VALUE_STORAGE_COLLECTION_NAME,
                 key_value_storage: Optional[storage.KeyValueStorage] = None,
                 map_storage: Optional[storage.MapStorage] = None,
                 wallet_storage: Optional[storage.WalletStorage] = None,
                 mongo_database: Optional[MongoDatabase] = None,
                 gateway_controller: Optional[controller.GatewayController] = None,
                 coin_max_handle_transaction_tries: int = DEFAULT_COIN_MAX_HANDLE_TRANSACTION_TRIES,
                 waves_max_handle_transaction_tries: int = DEFAULT_WAVES_MAX_HANDLE_TRANSACTION_TRIES,
                 coin_last_block_distance: int = DEFAULT_COIN_LAST_BLOCK_DISTANCE,
                 waves_last_block_distance: int = DEFAULT_WAVES_LAST_BLOCK_DISTANCE,
                 web_primary_color: str = DEFAULT_WEB_PRIMARY_COLOR) -> None:
        """
        Creates a new Gateway instance.

        :param coin_address_factory:
            Instance of an CoinAddressFactory that must be provided.
            Provides the functionality to create new addresses of the custom currency.

        :param coin_chain_query_service:
            This must be an instance of a CoinChainQuery service sub-class.
            It allows the gateway to perform certain queries on the Blockchain of the custom currency.

        :param gateway_waves_address_secret:
            Represents the address of the Gateway in Waves.

        :param coin_transaction_service:
            Implements the functionality to perform transactions in the custom currency.

        :param waves_node:
            Sets the Waves node that shall be used, for example: http://5.189.136.6:6869.

        :param waves_asset_id:
            The identifier for the Waves Asset that shall be used.

        :param gateway_owner_address:
            The coin address of the person who owns the Gateway; runs the server.
            A small amount of coins will be transferred to this address every time the Gateway performs
            a Waves transaction. The concrete amount is calculated based on the
            given percentage 'gateway_owner_waves_fee_percentage'.

        :param fee_service:

        :param only_one_transaction_receiver:
            Defines whether the custom cryptocurrency supports only one receiver of a transaction.

        :param waves_chain:
            Allows to configure the waves_chain to be used. This can either be 'mainnet' or 'testnet'.
            By default 'mainnet' will be used.

        :param custom_currency_name:
            Name that shall be used when something regarding the custom currency is displayed.
            Major use is in the Web App.

        :param logging_handlers:
            An array of logging.Handler implementations to use. By default a single StreamHandler
            is used with an DEBUG level.

        :param managed_loggers:
            Allows to process other logger inputs. Uses logging.getLogger with the given names to
            fetch those additional loggers.

        :param host
            The hostname to be used for the Web Server. Default is localhost.

        :param port
            The port to be used for the Web Server. Default is 5000.

        :param polling_delay_s:
            Optional parameter that allows to specify the delay in seconds between polling
            schedules of the block poller calls.

        :param wallet_storage_collection_name:
            May overwrite the collection name of the wallet storage in the MongoDB.
            This has no effect, if a custom storage implementation is used.

        :param map_storage_collection_name:
            May overwrite the collection name of the map storage in the MongoDB.
            This has no effect, if a custom storage implementation is used.

        :param key_value_storage_collection_name:
            May overwrite the collection name of the key_value storage in the MongoDB.
            This has no effect, if a custom storage implementation is used.

        :param key_value_storage:
            Is used to store key-value pairs. Defaults to an MongoDB implementation.

        :param map_storage:
            Stores the associations between custom currency addresses and Waves addresses.

        :param wallet_storage:
            Stores the secrets of the private keys created by the CoinAddressFactory.

        :param mongo_database:
            This is marked as Optional, but is required if no custom storage implementations are used.
            This must be a valid instance of a MongoDB database of the PyMongo framework.

        :param gateway_controller:
            This is an instance of the external API of the Gateway. For example,
            the Gateway offers the possibility to create an ReST interface by specifying a Flask instance.
            Please take a look at the flask parameter description for more information.
            The auto-generated ReST API uses this controller in the background to process the incoming requests.
            The default implementation should be sufficient for most use cases, but it may be overridden here.

        :param gateway_coin_address_secret:
            This is the major Gateway address for storing coins of the custom cryptocurrency.

        :param coin_last_block_distance:
            This may be used to prevent a check of the highest block.

        :param waves_last_block_distance:
            This may be used to prevent a check of the highest block.

        """
        self._injector = common.INJECTOR
        self._injector.overwrite(CoinAddressFactory, coin_address_factory)
        self._injector.overwrite(COIN_CHAIN_QUERY_SERVICE, coin_chain_query_service)
        self._injector.overwrite_if_exists(LOGGING_HANDLER_LIST, logging_handlers)
        self._injector.overwrite_if_exists(POLLING_DELAY_SECONDS, polling_delay_s)
        self._injector.overwrite_if_exists(PollingDelayConfig, polling_delay_config)
        self._injector.overwrite(CUSTOM_CURRENCY_NAME, custom_currency_name)
        self._injector.overwrite(GATEWAY_OWNER_ADDRESS, gateway_owner_address)
        self._injector.overwrite(WALLET_STORAGE_COLLECTION_NAME, wallet_storage_collection_name)
        self._injector.overwrite(MAP_STORAGE_COLLECTION_NAME, map_storage_collection_name)
        self._injector.overwrite(KEY_VALUE_STORAGE_COLLECTION_NAME, key_value_storage_collection_name)
        self._injector.overwrite(TRANSACTION_ATTEMPT_LIST_STORAGE_COLLECTION_NAME,
                                 transaction_attempt_list_storage_collection_name)
        self._injector.overwrite(GATEWAY_COIN_ADDRESS_SECRET, gateway_coin_address_secret)
        self._injector.overwrite(GATEWAY_COIN_ADDRESS, gateway_coin_address_secret.public)
        self._injector.overwrite(WAVES_NODE, waves_node)
        self._injector.overwrite(WAVES_ASSET_ID, waves_asset_id)
        self._injector.overwrite(WAVES_CHAIN, waves_chain)
        self._injector.overwrite(ONLY_ONE_TRANSACTION_RECEIVER, only_one_transaction_receiver)
        self._injector.overwrite_if_exists(COIN_TRANSACTION_WEB_LINK, coin_transaction_web_link)
        self._injector.overwrite_if_exists(WAVES_TRANSACTION_WEB_LINK, waves_transaction_web_link)
        self._injector.overwrite_if_exists(COIN_ADDRESS_WEB_LINK, coin_address_web_link)
        self._injector.overwrite_if_exists(WAVES_ADDRESS_WEB_LINK, waves_address_web_link)
        self._injector.overwrite(COIN_ADDRESS_VALIDATION_SERVICE, coin_address_validation_service)
        self._injector.overwrite(ATTEMPT_LIST_MAX_COMPLETION_TRIES, attempt_list_max_completion_tries)
        self._injector.overwrite(GATEWAY_WAVES_ADDRESS_SECRET, gateway_waves_address_secret)
        self._injector.overwrite(GATEWAY_WAVES_ADDRESS, gateway_waves_address_secret.public)
        self._injector.overwrite_if_exists(MongoDatabase, mongo_database)
        self._injector.overwrite_if_exists(storage.KeyValueStorage, key_value_storage)
        self._injector.overwrite_if_exists(storage.MapStorage, map_storage)
        self._injector.overwrite_if_exists(storage.WalletStorage, wallet_storage)
        self._injector.overwrite_if_exists(COIN_INTEGER_CONVERTER_SERVICE, coin_integer_converter_service)
        self._injector.overwrite_if_exists(ASSET_INTEGER_CONVERTER_SERVICE, asset_integer_converter_service)
        self._injector.overwrite(service.FeeService, fee_service)
        self._injector.overwrite(COIN_TRANSACTION_SERVICE, coin_transaction_service)
        self._injector.overwrite_if_exists(controller.GatewayController, gateway_controller)
        self._injector.overwrite(NUM_ATTEMPT_LIST_WORKERS, num_attempt_list_workers)
        self._injector.overwrite(GATEWAY_HOST, host)
        self._injector.overwrite(GATEWAY_PORT, port)
        self._injector.overwrite(FLASK_NAME, Gateway.DEFAULT_FLASK_NAME)
        self._injector.overwrite(COIN_MAX_HANDLE_TRANSACTION_TRIES, coin_max_handle_transaction_tries)
        self._injector.overwrite(WAVES_MAX_HANDLE_TRANSACTION_TRIES, waves_max_handle_transaction_tries)
        self._injector.overwrite(COIN_LAST_BLOCK_DISTANCE, coin_last_block_distance)
        self._injector.overwrite(WAVES_LAST_BLOCK_DISTANCE, waves_last_block_distance)
        self._injector.overwrite(WEB_PRIMARY_COLOR, web_primary_color)

        self._init_managed_loggers(managed_loggers)

        self._application_service = self._injector.get(
            service.GatewayApplicationService)  # type: service.GatewayApplicationService

        self._flask_rest_controller = self._injector.get(controller.FlaskRestController)
        self._logging_service = self._injector.get(
            service.GatewayLoggingConfigurationService)  # type: service.GatewayLoggingConfigurationService

        self.set_log_level = self._logging_service.set_log_level
        self.run = self._application_service.run



"""
ChainQueryService
"""

from abc import ABC, abstractmethod
from typing import List, Union, Optional

from decimal import Decimal

from waves_gateway.common import InjectionToken
from waves_gateway.model import Transaction


class ChainQueryService(ABC):
    """
    Provides read operations on the chain of the custom cryptocurrency
    """

    @abstractmethod
    def get_transactions_of_block_at_height(self, height: int) -> List[Transaction]:
        """
        Returns the transactions of the block at the given height.
        """
        pass

    @abstractmethod
    def get_height_of_highest_block(self) -> int:
        """
        Returns the newest block of the chain.
        """
        pass

    def get_coin_receiver_address_from_transaction(self, transaction: str):
        """
        Returns the raw (unparsed) receiver address string.
        """
        raise NotImplementedError()

    def get_transaction_by_tx(self, tx: str) -> Optional[Transaction]:
        """
        Returns the associated transaction if it exists. This functionality is optional.
        If the given identifier is invalid, the function should raise an InvalidTransactionIdentifier error.
        Otherwise, the return value None means that the transaction exists, but is not relevant for the Gateway.
        """
        raise NotImplementedError()



"""
MappingEntrySerializer
"""
from waves_gateway.common import Injectable
from waves_gateway.model import MappingEntry


@Injectable()
class MappingEntrySerializer(object):
    """
    Defines how a MappingEntry can be serialized and deserialized.
    """

    # noinspection PyMethodMayBeStatic
    def as_dict(self, entry: MappingEntry) -> dict:
        """
        Converts a mapping entry into a dictionary containing all attributes.
        """
        res = dict()

        res[MappingEntry.DICT_WAVES_KEY] = entry.waves_address
        res[MappingEntry.DICT_COIN_KEY] = entry.coin_address

        return res



"""
BlockHeightStorageProxy
"""

from abc import ABC, abstractmethod
from typing import Optional


class BlockHeightStorageProxy(ABC):
    """
    Storage for the Block height. This storage is an abstraction layer for the polling service
    and is not meant to be provided by the framework user.
    Internally, it forwards the requests to the KeyValueStorage.
    """

    @abstractmethod
    def get_last_checked_block_height(self) -> Optional[int]:
        """
        Returns the stored last checked block height or None if it was never set.
        """
        pass

    @abstractmethod
    def set_last_checked_block_height(self, block_height: int) -> None:
        """
        Overwrites the last checked block height.
        """



"""
SecretService
"""

import pywaves  # type: ignore

from waves_gateway.common import Injectable, GATEWAY_COIN_ADDRESS_SECRET, GATEWAY_PYWAVES_ADDRESS
from waves_gateway.storage import WalletStorage
from waves_gateway.model import KeyPair


@Injectable(deps=[WalletStorage, GATEWAY_COIN_ADDRESS_SECRET, GATEWAY_PYWAVES_ADDRESS])
class SecretService(object):
    """
    Forwards to the specific secret if any.
    Can be used to query any secret that the Gateway stores.
    """

    def __init__(self, wallet_storage: WalletStorage, gateway_coin_address_secret: KeyPair,
                 gateway_pywaves_address: pywaves.Address) -> None:
        self._wallet_storage = wallet_storage
        self._coin_gateway_address_secret = gateway_coin_address_secret
        self._gateway_pywaves_address = gateway_pywaves_address

    def get_secret_by_address(self, currency: str, address: str):
        if currency == "waves" and self._gateway_pywaves_address.address == address:
            return self._gateway_pywaves_address.privateKey
        elif currency == "coin":
            if self._coin_gateway_address_secret.public == address:
                return self._coin_gateway_address_secret.secret
            else:
                key_pair = self._wallet_storage.get_secret_by_public_address(address)
                if key_pair is not None:
                    return key_pair.secret
        return None



"""
GatewayControllerConverterProxyImpl
"""

from typing import Optional, List
from waves_gateway.common import map_array, Injectable
from waves_gateway.service import AttemptListConverterService

from waves_gateway.model import AttemptListTrigger, TransactionAttemptList, AttemptListQuery

from .gateway_controller import GatewayController


@Injectable(deps=[AttemptListConverterService, GatewayController])
class GatewayControllerConverterProxyImpl(GatewayController):
    """
    Reverts the amount conversion which the Gateway internally uses before overhanding the results
    to an external API.
    """

    def check_coin_transaction(self, tx: str) -> None:
        self._gateway_controller.check_coin_transaction(tx)

    def check_waves_transaction(self, tx: str) -> None:
        self._gateway_controller.check_waves_transaction(tx)

    def validate_waves_address(self, address: str) -> bool:
        return self._gateway_controller.validate_waves_address(address)

    def query_attempt_lists(self, query: AttemptListQuery) -> List[TransactionAttemptList]:
        results = self._gateway_controller.query_attempt_lists(query)
        return map_array(lambda x: self._converter.revert_attempt_list_conversion(x), results)

    def get_attempt_list_by_id(self, attempt_list_id: str) -> Optional[TransactionAttemptList]:
        result = self._gateway_controller.get_attempt_list_by_id(attempt_list_id)
        if result is None:
            return None
        else:
            return self._converter.revert_attempt_list_conversion(result)

    def __init__(self, attempt_list_converter_service: AttemptListConverterService,
                 gateway_controller: GatewayController) -> None:
        self._converter = attempt_list_converter_service
        self._gateway_controller = gateway_controller

    def get_attempt_list_by_trigger(self, trigger: AttemptListTrigger) -> Optional[TransactionAttemptList]:
        return self._converter.revert_attempt_list_conversion(
            self._gateway_controller.get_attempt_list_by_trigger(trigger))

    def create_address(self, waves_address: str) -> str:
        return self._gateway_controller.create_address(waves_address)



"""
Waves Gateway Library
========================

Contains tools to create a new Waves Cryptocurrency Gateway.
It is recommended to call gevent.monkey.patch_all() at the very beginning of the application
as this library uses gevent for non-blocking io.
"""

import gevent.monkey
gevent.monkey.patch_all()

import gevent.hub
gevent.hub.Hub.NOT_ERROR = (Exception, )

from waves_gateway.factory import CoinAddressFactory
from waves_gateway.model import KeyPair, \
    TransactionAttemptReceiver, \
    TransactionAttempt, Transaction, TransactionReceiver, TransactionSender, PollingDelayConfig, GatewayConfigFile
from .gateway import Gateway
from waves_gateway.service import ChainQueryService, TransactionService, IntegerConverterService, \
    FeeService, ConstantFeeServiceImpl, AddressValidationService, COIN_INTEGER_CONVERTER_SERVICE, COIN_TRANSACTION_SERVICE, COIN_CHAIN_QUERY_SERVICE, COIN_ADDRESS_VALIDATION_SERVICE, GatewayConfigParser, GatewayApplicationService
from waves_gateway.storage import WalletStorage, MapStorage, KeyValueStorage
from waves_gateway.common import convert_to_int, convert_to_decimal, ProxyGuard, InvalidTransactionIdentifier, InjectionToken, Injectable, Factory, INJECTOR, Injector, InjectorError, Token, COIN_NODE, CUSTOM_CURRENCY_NAME, WAVES_ASSET_ID, ProxyFactory

from typing import NewType

CoinAddress = NewType('CoinAddress', str)
WavesAddress = NewType('WavesAddress', str)

WavesAddressSecret = NewType('WavesAddressSecret', KeyPair)
CoinAddressSecret = NewType('CoinAddressSecret', KeyPair)

CoinTransactionIdentifier = NewType('CoinTransactionIdentifier', str)
WavesTransactionIdentifier = NewType('WavesTransactionIdentifier', str)

CoinBlockHeight = NewType('CoinBlockHeight', int)
WavesBlockHeight = NewType('WavesBlockHeight', int)

WavesTransaction = NewType('WavesTransaction', Transaction)
CoinTransaction = NewType('CoinTransaction', Transaction)

CoinAmount = NewType('CoinAmount', int)



"""
PollingState
"""

from typing import Dict


class PollingTransactionState(object):
    """
    Stores information about a processed transaction.
    """

    DICT_TRIES_KEY = 'tries'
    DICT_OK_KEY = 'ok'

    def __init__(self, ok: bool = False, tries: int = 0) -> None:
        self._ok = ok
        self._tries = tries

    @property
    def tries(self) -> int:
        return self._tries

    @property
    def ok(self) -> bool:
        return self._ok

    def increment_tries(self) -> None:
        self._tries = self._tries + 1

    def mark_as_done(self):
        self._ok = True

    def __eq__(self, other):
        return self.tries == other.tries and self.ok == other.ok


class PollingState(object):
    """
    Saves information about the last poller execution.
    """

    DICT_TRANSACTION_MAP_KEY = 'transaction_map'

    def __init__(self, transaction_map: Dict[str, PollingTransactionState] = None) -> None:

        if transaction_map is None:
            self._transaction_map = dict()  # type: Dict[str, PollingTransactionState]
        else:
            self._transaction_map = transaction_map

    def __eq__(self, other):
        return self.transaction_map == other.transaction_map

    @property
    def transaction_map(self) -> Dict[str, PollingTransactionState]:
        return self._transaction_map



"""
PollingDelayConfig
"""
from typing import Any


class PollingDelayConfig(object):
    """
    Summarized configuration for the polling_delay settings in the Gateway Application.
    """

    DEFAULT_MIN_TRANSACTION_POLLING_DELAY_S = 0.0
    DEFAULT_MAX_TRANSACTION_POLLING_DELAY_S = 60.0
    DEFAULT_MIN_ATTEMPT_LIST_WORKER_DELAY_S = 0.1
    DEFAULT_MAX_ATTEMPT_LIST_WORKER_DELAY_S = 60.0

    def __init__(self,
                 coin_min_polling_delay_s: float = DEFAULT_MIN_TRANSACTION_POLLING_DELAY_S,
                 coin_max_polling_delay_s: float = DEFAULT_MAX_TRANSACTION_POLLING_DELAY_S,
                 waves_min_polling_delay_s: float = DEFAULT_MIN_TRANSACTION_POLLING_DELAY_S,
                 waves_max_polling_delay_s: float = DEFAULT_MAX_TRANSACTION_POLLING_DELAY_S,
                 attempt_list_worker_min_polling_delay_s: float = DEFAULT_MIN_ATTEMPT_LIST_WORKER_DELAY_S,
                 attempt_list_worker_max_polling_delay_s: float = DEFAULT_MAX_ATTEMPT_LIST_WORKER_DELAY_S) -> None:
        self._coin_polling_delay_s_min = coin_min_polling_delay_s
        self._coin_polling_delay_s_max = coin_max_polling_delay_s
        self._waves_polling_delay_s_min = waves_min_polling_delay_s
        self._waves_polling_delay_s_max = waves_max_polling_delay_s
        self._attempt_list_worker_min_polling_delay_s = attempt_list_worker_min_polling_delay_s
        self._attempt_list_worker_max_polling_delay_s = attempt_list_worker_max_polling_delay_s

    @staticmethod
    def from_single_polling_delay(polling_delay_s: float) -> Any:
        return PollingDelayConfig(
            coin_min_polling_delay_s=polling_delay_s,
            coin_max_polling_delay_s=polling_delay_s,
            waves_min_polling_delay_s=polling_delay_s,
            waves_max_polling_delay_s=polling_delay_s,
            attempt_list_worker_min_polling_delay_s=polling_delay_s,
            attempt_list_worker_max_polling_delay_s=polling_delay_s)

    @property
    def waves_max_polling_delay_s(self) -> float:
        return self._waves_polling_delay_s_max

    @property
    def waves_min_polling_delay_s(self) -> float:
        return self._waves_polling_delay_s_min

    @property
    def coin_min_polling_delay_s(self) -> float:
        return self._coin_polling_delay_s_min

    @property
    def coin_max_polling_delay_s(self) -> float:
        return self._coin_polling_delay_s_max

    @property
    def attempt_list_worker_min_polling_delay_s(self) -> float:
        return self._attempt_list_worker_min_polling_delay_s

    @property
    def attempt_list_worker_max_polling_delay_s(self) -> float:
        return self._attempt_list_worker_max_polling_delay_s



"""
GatewayApplicationService
"""

from logging import Logger

import signal
from typing import List

import flask
import gevent
import gevent.pywsgi
import gevent.pool

from waves_gateway.model import PollingDelayConfig
from waves_gateway.common import Injectable, NUM_ATTEMPT_LIST_WORKERS, GATEWAY_HOST, GATEWAY_PORT
from .gateway_validation_service import GatewayValidationService
from .transaction_polling_service import TransactionPollingService
from .attempt_list_worker import AttemptListWorker
from .transaction_attempt_list_service import TransactionAttemptListService
from .pending_attempt_list_selector_service import PendingAttemptListSelectorService
from .token import COIN_TRANSACTION_POLLING_SERVICE, WAVES_TRANSACTION_POLLING_SERVICE


@Injectable(deps=[
    GatewayValidationService, Logger, COIN_TRANSACTION_POLLING_SERVICE, WAVES_TRANSACTION_POLLING_SERVICE,
    NUM_ATTEMPT_LIST_WORKERS, TransactionAttemptListService, PendingAttemptListSelectorService, PollingDelayConfig,
    flask.Flask, GATEWAY_HOST, GATEWAY_PORT
])
class GatewayApplicationService(object):
    """
    Implements the run method of the Gateway. This functionality was moved from the main Gateway class to here
    to simplify the main Gateway class. The run method starts all pollers.
    """

    def __init__(self, validation_service: GatewayValidationService, logger: Logger,
                 coin_transaction_polling_service: TransactionPollingService,
                 waves_transaction_polling_service: TransactionPollingService, num_attempt_list_workers: int,
                 attempt_list_service: TransactionAttemptListService,
                 attempt_list_selector: PendingAttemptListSelectorService, polling_delay_config: PollingDelayConfig,
                 flask: flask.Flask, host: str, port: int) -> None:
        self._validation_service = validation_service
        self._logger = logger.getChild(GatewayApplicationService.__name__)
        self._coin_transaction_polling_service = coin_transaction_polling_service
        self._waves_transaction_polling_service = waves_transaction_polling_service
        self._num_attempt_list_workers = num_attempt_list_workers
        self._attempt_list_service = attempt_list_service
        self._attempt_list_selector = attempt_list_selector
        self._polling_delay_config = polling_delay_config
        self._host = host
        self._port = port
        self._flask = flask

    def _create_attempt_list_workers(self) -> List[AttemptListWorker]:
        tasks = list()  # type: List[AttemptListWorker]

        for i in range(0, self._num_attempt_list_workers):
            task = AttemptListWorker(
                attempt_list_service=self._attempt_list_service,
                attempt_list_selector=self._attempt_list_selector,
                logger=self._logger,
                worker_id=i,
                max_polling_delay_s=self._polling_delay_config.attempt_list_worker_max_polling_delay_s,
                min_polling_delay_s=self._polling_delay_config.attempt_list_worker_min_polling_delay_s)

            tasks.append(task)

        return tasks

    def run(self) -> None:
        """
        Starts all poller instances.
        After that, polling is performed in regular intervals specified by the polling_delay_ms property.
        By default this function blocks the current thread.
        """
        self._validation_service.validate_all_addresses()

        self._logger.info("Gateway Application started")

        task_group = gevent.pool.Group()

        gevent.signal(signal.SIGINT, self._coin_transaction_polling_service.cancel)
        gevent.signal(signal.SIGINT, self._waves_transaction_polling_service.cancel)

        task_group.start(self._coin_transaction_polling_service)
        task_group.start(self._waves_transaction_polling_service)

        attempt_list_workers = self._create_attempt_list_workers()

        for worker in attempt_list_workers:
            gevent.signal(signal.SIGINT, worker.cancel)
            task_group.start(worker)

        http = gevent.pywsgi.WSGIServer(
            (self._host, self._port), self._flask, log=gevent.pywsgi.LoggingLogAdapter(self._logger.getChild('pywsgi')))

        gevent.signal(signal.SIGINT, http.close)

        self._logger.info('Listening on %s:%s', self._host, str(self._port))

        http.serve_forever()

        task_group.join(raise_error=True)



"""
TransactionAttempt
"""

from typing import List, Union

from decimal import Decimal


class TransactionAttemptReceiver(object):
    """
    Defines the amount that shall be transferred to a specific receiver address.
    """
    DICT_ADDRESS_KEY = 'address'
    DICT_AMOUNT_KEY = 'amount'

    def __init__(self, address: str, amount: Union[int, Decimal]) -> None:
        self._address = address
        self._amount = amount

    @property
    def address(self) -> str:
        return self._address

    @property
    def amount(self) -> Union[int, Decimal]:
        return self._amount

    def __eq__(self, other):
        if not isinstance(other, TransactionAttemptReceiver):
            return False
        else:
            res = True
            res = res and self._address == other.address
            res = res and self._amount == other.amount
            return res


class TransactionAttempt(object):
    """
    An TransactionAttempt defines a planned transaction. This is a transaction
    that may not have been performed, yet. But is meant to be performed in the future.
    """
    DICT_FEE = "fee"
    DICT_SENDER = "sender"
    DICT_RECEIVERS = "receivers"
    DICT_CURRENCY = "currency"

    def __init__(
            self,
            sender: str,
            receivers: List[TransactionAttemptReceiver],
            fee: Union[int, Decimal],
            currency: str,
    ) -> None:
        self._sender = sender
        self._receivers = receivers
        self._fee = fee
        self._currency = currency

    @property
    def fee(self) -> Union[int, Decimal]:
        return self._fee

    @property
    def sender(self) -> str:
        return self._sender

    @property
    def receivers(self) -> List[TransactionAttemptReceiver]:
        return self._receivers

    @property
    def currency(self) -> str:
        return self._currency

    def overall_amount(self) -> Union[int, Decimal]:
        amount = 0

        for receiver in self._receivers:
            amount = amount + receiver.amount  # type: ignore

        return amount

    def __eq__(self, other):
        res = True
        res = res and self.receivers == other.receivers
        res = res and self.sender == other.sender
        res = res and self.fee == other.fee
        res = res and self.currency == other.currency

        return res



"""
AssetTransactionServiceConverterProxyImpl
"""

from typing import Optional, cast

from waves_gateway.common import Injectable
from waves_gateway.model import Transaction, TransactionAttempt
from waves_gateway.service.token import COIN_TRANSACTION_SERVICE, COIN_INTEGER_CONVERTER_SERVICE, \
    ASSET_INTEGER_CONVERTER_SERVICE

from .attempt_list_converter_service import AttemptListConverterService

from .transaction_service import TransactionService
from .integer_converter_service import IntegerConverterService
from .asset_transaction_service_impl import AssetTransactionServiceImpl
from .token import COIN_TRANSACTION_SERVICE_CONVERTER_PROXY, WAVES_TRANSACTION_SERVICE_CONVERTER_PROXY


@Injectable(
    COIN_TRANSACTION_SERVICE_CONVERTER_PROXY,
    deps=[COIN_INTEGER_CONVERTER_SERVICE, COIN_TRANSACTION_SERVICE, AttemptListConverterService])
@Injectable(
    WAVES_TRANSACTION_SERVICE_CONVERTER_PROXY,
    deps=[ASSET_INTEGER_CONVERTER_SERVICE, AssetTransactionServiceImpl, AttemptListConverterService])
class TransactionServiceConverterProxyImpl(TransactionService):
    """
    Reverts the amount conversion before overhanding the TransactionAttempt instance
    to the given TransactionService.
    """

    def __init__(self, integer_converter_service: IntegerConverterService, transaction_service: TransactionService,
                 attempt_list_converter_service: AttemptListConverterService) -> None:
        self._integer_converter_service = integer_converter_service
        self._transaction_service = transaction_service
        self._attempt_list_converter_service = attempt_list_converter_service

    def send_coin(self, attempt: TransactionAttempt, secret: Optional[str]) -> Transaction:
        converted_attempt = self._attempt_list_converter_service.revert_attempt_conversion(attempt)
        transaction = self._transaction_service.send_coin(converted_attempt, secret)
        return self._integer_converter_service.convert_transaction_to_int(transaction)



"""
KeyPairSerializer
"""

from waves_gateway.common import ParseError, Injectable
from waves_gateway.model import KeyPair


@Injectable()
class KeyPairSerializer(object):
    """
    Can be used to serialize and deserialize a KeyPair derived alias such as an
    AccountSecret instance or an AddressSecret.
    """

    # noinspection PyMethodMayBeStatic
    def as_dict(self, key_pair: KeyPair) -> dict:
        """
        Transforms a KeyPair into a dictionary that contains all attributes of the KeyPair.
        After that, the dict may be persisted and deserialized later.
        For example, the dict representation may be converted into JSON using the standard json library.
        """
        res = dict()

        res[KeyPair.DICT_PUBLIC_KEY] = key_pair.public
        res[KeyPair.DICT_SECRET_KEY] = key_pair.secret

        return res

    # noinspection PyMethodMayBeStatic
    def from_dict(self, data: dict) -> KeyPair:
        """
        Tries to reconstruct a KeyPair from the given dictionary.
        This may fail if the dictionary does not define the necessary properties.
        In this case a ParseError is thrown.
        """

        if KeyPair.DICT_PUBLIC_KEY not in data:
            raise ParseError("required key '" + KeyPair.DICT_PUBLIC_KEY + "' is missing")

        if KeyPair.DICT_SECRET_KEY not in data:
            raise ParseError("required key '" + KeyPair.DICT_SECRET_KEY + "' is missing")

        return KeyPair(public=data[KeyPair.DICT_PUBLIC_KEY], secret=data[KeyPair.DICT_SECRET_KEY])



"""
WavesPollingStateStorageProxyImpl
"""

from waves_gateway.common import Injectable

from waves_gateway.model import PollingState
from .polling_state_storage_proxy import PollingStateStorageProxy
from .key_value_storage import KeyValueStorage


@Injectable(deps=[KeyValueStorage])
class WavesPollingStateStorageProxyImpl(PollingStateStorageProxy):
    """Forwards polling_state requests to the waves methods of KeyValueStorage."""

    def __init__(self, key_value_storage: KeyValueStorage) -> None:
        self._storage = key_value_storage

    def set_polling_state(self, state: PollingState) -> None:
        self._storage.set_waves_polling_state(state)

    def get_polling_state(self) -> PollingState:
        return self._storage.get_waves_polling_state()



"""
Transaction
"""

from typing import List, Union

from decimal import Decimal


class TransactionSender(object):
    """
    Stores information about the sender of an transaction.
    """
    DICT_ADDRESS_KEY = 'address'

    def __init__(self, address: str) -> None:
        self._address = address

    @property
    def address(self) -> str:
        return self._address

    def __eq__(self, other):
        if not isinstance(other, TransactionSender):
            return False
        else:
            res = True
            res = res and self._address == other.address
            return res


class TransactionReceiver(object):
    """
    Defines the amount which the transaction spends to a specific address.
    """
    DICT_ADDRESS_KEY = 'address'
    DICT_AMOUNT_KEY = 'amount'

    def __init__(self, address: str, amount: Union[int, Decimal]) -> None:
        self._address = address
        self._amount = amount

    @property
    def address(self) -> str:
        return self._address

    @property
    def amount(self) -> Union[int, Decimal]:
        return self._amount

    def __eq__(self, other):
        if not isinstance(other, TransactionReceiver):
            return False
        else:
            res = True
            res = res and self._address == other.address
            res = res and self._amount == other.amount
            return res


class Transaction(object):
    """
    Represents a transaction in any coin (including Waves).
    The sender may be any kind of public identifier that can be written as a string.
    In this package, this is the public part of some KeyPair derived class.
    """

    DICT_RECEIVERS_KEY = 'receivers'
    DICT_TX_KEY = 'tx'

    def __init__(self, tx: str, receivers: List[TransactionReceiver], senders: List[TransactionSender] = None) -> None:
        self._tx = tx
        self._receivers = receivers
        self._senders = senders

    @property
    def tx(self) -> str:
        return self._tx

    @property
    def receivers(self) -> List[TransactionReceiver]:
        return self._receivers

    @property
    def senders(self) -> List[TransactionSender]:
        return self._senders

    def __eq__(self, other):
        if not isinstance(other, Transaction):
            return False
        else:
            res = True
            res = res and self.receivers == other.receivers
            res = res and self.tx == other.tx
            # TODO: add senders here also???
            return res

    def __str__(self):
        return "Transaction(" + str(self.tx) + ")"



"""
GatewayValidationService
"""

from waves_gateway.model import KeyPair
from .address_validation_service import AddressValidationService
from waves_gateway.common import Injectable, InvalidConfigError, CUSTOM_CURRENCY_NAME, GATEWAY_OWNER_ADDRESS, \
    GATEWAY_COIN_ADDRESS_SECRET, GATEWAY_WAVES_ADDRESS_SECRET
from .waves_address_validation_service import WavesAddressValidationService
from .token import COIN_ADDRESS_VALIDATION_SERVICE


@Injectable(deps=[
    WavesAddressValidationService, COIN_ADDRESS_VALIDATION_SERVICE, GATEWAY_WAVES_ADDRESS_SECRET, GATEWAY_OWNER_ADDRESS,
    CUSTOM_CURRENCY_NAME, GATEWAY_COIN_ADDRESS_SECRET
])
class GatewayValidationService(object):
    """
    Implements validation methods that may be used the Gateway application. The validate_all_addresses is meant to
    be called right on initialization to prevent the usage of misleading addresses.
    """

    def __init__(self, waves_address_validation_service: AddressValidationService,
                 coin_address_validation_service: AddressValidationService, gateway_waves_address_secret: KeyPair,
                 gateway_owner_address: str, custom_currency_name: str, gateway_coin_address_secret: KeyPair) -> None:
        self._waves_address_validation_service = waves_address_validation_service
        self._coin_address_validation_service = coin_address_validation_service
        self._gateway_waves_address_secret = gateway_waves_address_secret
        self._gateway_owner_address = gateway_owner_address
        self._custom_currency_name = custom_currency_name
        self._gateway_coin_address_secret = gateway_coin_address_secret

    def validate_all_addresses(self):
        if not self._waves_address_validation_service.validate_address(self._gateway_waves_address_secret.public):
            raise InvalidConfigError("The public part of the given waves address "
                                     "KeyPair is not a valid waves address.")

        if not self._coin_address_validation_service.validate_address(self._gateway_owner_address):
            raise InvalidConfigError(
                "The gateway owner address is not a valid " + self._custom_currency_name + " address.")

        if not self._coin_address_validation_service.validate_address(self._gateway_coin_address_secret.public):
            raise InvalidConfigError(
                "The public part of the coin address is not a valid " + self._custom_currency_name + " address.")



"""
IntegerConverterService
"""

from abc import ABC
from typing import Union, cast
from copy import deepcopy

from decimal import Decimal

from waves_gateway.model import Transaction, TransactionReceiver


class IntegerConverterService(ABC):
    """
    Defines the requires functionality for an IntegerConverterService.
    It must allow the conversion from a float value to an int value and backwards.
    By default, it does not perform any conversion.
    """

    # noinspection PyMethodMayBeStatic
    def revert_amount_conversion(self, amount: int) -> Union[int, Decimal]:
        """Returns the argument by default (if not overwritten)."""
        return amount

    # noinspection PyMethodMayBeStatic
    def convert_amount_to_int(self, amount: Union[int, Decimal]) -> int:
        """Returns the argument by default (if not overwritten)."""
        return cast(int, amount)

    def safely_convert_to_int(self, amount: Union[int, Decimal]) -> int:
        """
        Tests the result of convert_amount_to_int if it has the type int.
        If not, it raises an TypeError.
        """
        converted = self.convert_amount_to_int(amount)

        if isinstance(converted, int):
            return converted
        else:
            raise TypeError('value must be an integer; value was ' + str(converted))

    def convert_transaction_to_int(self, transaction: Transaction) -> Transaction:
        res = deepcopy(transaction)

        for i in range(0, len(res.receivers)):
            receiver = res.receivers[i]
            res.receivers[i] = TransactionReceiver(receiver.address, self.safely_convert_to_int(receiver.amount))

        return res

    def revert_transaction_conversion(self, transaction: Transaction) -> Transaction:
        res = deepcopy(transaction)

        for i in range(0, len(res.receivers)):
            receiver = res.receivers[i]
            res.receivers[i] = TransactionReceiver(receiver.address,
                                                   self.revert_amount_conversion(cast(int, receiver.amount)))

        return res



"""
CoinTransactionConsumerImpl
"""

from logging import Logger
from typing import List, cast

import datetime
import pywaves  # type: ignore

from waves_gateway.common import MappingNotFoundForCoinAddress, \
    MultipleGatewayReceiversError, Injectable, GATEWAY_OWNER_ADDRESS, GATEWAY_COIN_ADDRESS, \
    ONLY_ONE_TRANSACTION_RECEIVER, GATEWAY_WAVES_ADDRESS, GATEWAY_PYWAVES_ADDRESS
from waves_gateway.model import Transaction, TransactionReceiver, \
    TransactionAttempt, TransactionAttemptReceiver, TransactionAttemptList, \
    AttemptListTrigger, TransactionSender
from waves_gateway.storage import MapStorage, TransactionAttemptListStorage
from .fee_service import FeeService
from .transaction_attempt_list_service import TransactionAttemptListService
from .transaction_consumer import TransactionConsumer
from .fee_service_converter_proxy_impl import FeeServiceConverterProxyImpl


@Injectable(deps=[
    TransactionAttemptListService, GATEWAY_COIN_ADDRESS, MapStorage, Logger, GATEWAY_WAVES_ADDRESS,
    GATEWAY_OWNER_ADDRESS, FeeServiceConverterProxyImpl, ONLY_ONE_TRANSACTION_RECEIVER, GATEWAY_PYWAVES_ADDRESS,
    TransactionAttemptListStorage
])
class CoinTransactionConsumerImpl(TransactionConsumer):
    """
    Handles fetched transactions of the custom currency.
    """

    def __init__(self, attempt_service: TransactionAttemptListService, gateway_coin_address: str,
                 map_storage: MapStorage, logger: Logger, gateway_waves_address: str, gateway_owner_address: str,
                 fee_service: FeeService, only_one_transaction_receiver: bool, gateway_pywaves_address: pywaves.Address,
                 attempt_list_storage: TransactionAttemptListStorage) -> None:
        self._attempt_service = attempt_service
        self._gateway_coin_address = gateway_coin_address
        self._map_storage = map_storage
        self._logger = logger.getChild(self.__class__.__name__)
        self._gateway_waves_address = gateway_waves_address
        self._gateway_owner_address = gateway_owner_address
        self._fee_service = fee_service
        self._only_one_transaction_receiver = only_one_transaction_receiver
        self._gateway_pywaves_address = gateway_pywaves_address
        self._attempt_list_storage = attempt_list_storage

    def _is_gateway_managed_coin_address(self, address: str) -> bool:
        """
        Checks if there is a mapping entry belonging to the given address.
        """
        return self._map_storage.coin_address_exists(address)

    def _is_gateway_coin_holder(self, address: str):
        return address == self._gateway_coin_address

    def filter_transaction(self, transaction: Transaction) -> bool:
        """
        Returns whether the transaction is addressed to a gateway managed address.
        This is the case if at least one of the receivers is a gateway managed address.
        But, not the main gateway address.
        """
        if self._attempt_list_storage.gateway_transaction_exists(transaction.tx):
            return False
        else:
            return len(self._filter_receivers(transaction)) > 0

    def _filter_receivers(self, transaction: Transaction) -> List[int]:
        """
        Filters a given list of transaction receivers for those ones that are gateway managed, but not the
        main gateway coin holder. Those shall be further processed.
        """
        res = list()

        for i in range(0, len(transaction.receivers)):
            address = transaction.receivers[i].address

            if self._is_gateway_managed_coin_address(address) and not self._is_gateway_coin_holder(address):
                attempt_list = self._attempt_list_storage.find_by_trigger(
                    AttemptListTrigger(tx=transaction.tx, currency="coin", receiver=i))

                if attempt_list is None or not attempt_list.is_complete():
                    res.append(i)

        return res

    def _handle_transaction_receiver(self,
                                     tx: str,
                                     receiver: TransactionReceiver,
                                     index: int,
                                     senders: List[TransactionSender] = None):

        # ---------- Pre-Calculation ----------

        # the gateway address that coins were transferred to
        gateway_managed_address = receiver.address

        # the custom currency fee to be used
        coin_fee = cast(int, self._fee_service.get_coin_fee())

        # the waves fee to be used; not used in the calculation
        waves_fee = cast(int, self._fee_service.get_waves_fee())

        # the fee for the gateway owner; service costs
        gateway_fee = cast(int, self._fee_service.get_gateway_fee())

        # the amount of coins that were received by this receiver
        received_amount = cast(int, receiver.amount)

        # the waves address to receive the amount of coins
        receiver_waves_address = self._map_storage.get_waves_address_by_coin_address(gateway_managed_address)

        # the amount of coins to be transferred after applying the fees
        if self._only_one_transaction_receiver:
            amount_after_fees = received_amount - 2 * coin_fee - gateway_fee
        else:
            amount_after_fees = received_amount - coin_fee - gateway_fee

        # ---------- Pre-Check ----------

        # transaction must have minimum amount to be handled
        if amount_after_fees <= 0:
            self._logger.warn("Received transaction %s with an amount of %s, but it is less than the at least "
                              "required amount. Will be skipped, but marked as processed.", tx, receiver.amount)
            return

        # no mapping was found for some reason; this in an internal error as this should
        # already be prevented by the filter method
        if receiver_waves_address is None:
            raise MappingNotFoundForCoinAddress(gateway_managed_address)

        attempt_list = self._attempt_list_storage.find_by_trigger(
            AttemptListTrigger(tx=tx, currency="coin", receiver=index))

        if attempt_list is None:

            attempts = list()

            if self._only_one_transaction_receiver:
                attempts.append(
                    TransactionAttempt(
                        sender=gateway_managed_address,
                        currency="coin",
                        fee=coin_fee,
                        receivers=[
                            TransactionAttemptReceiver(address=self._gateway_coin_address, amount=amount_after_fees),
                        ]))

                attempts.append(
                    TransactionAttempt(
                        sender=gateway_managed_address,
                        currency="coin",
                        fee=coin_fee,
                        receivers=[TransactionAttemptReceiver(address=self._gateway_owner_address,
                                                              amount=gateway_fee)]))
            else:
                attempts.append(
                    TransactionAttempt(
                        sender=gateway_managed_address,
                        currency="coin",
                        fee=coin_fee,
                        receivers=[
                            TransactionAttemptReceiver(address=self._gateway_coin_address, amount=amount_after_fees),
                            TransactionAttemptReceiver(address=self._gateway_owner_address, amount=gateway_fee)
                        ]))

            attempts.append(
                TransactionAttempt(
                    sender=self._gateway_pywaves_address.address,
                    fee=waves_fee,
                    currency="waves",
                    receivers=[TransactionAttemptReceiver(address=receiver_waves_address, amount=amount_after_fees)]))

            trigger = AttemptListTrigger(tx, index, "coin", senders=senders)
            attempt_list = TransactionAttemptList(
                trigger, attempts, last_modified=datetime.datetime.utcnow(), created_at=datetime.datetime.utcnow())
            self._attempt_list_storage.safely_save_attempt_list(attempt_list)
            self._logger.info('Created new attempt list %s', str(attempt_list.attempt_list_id))

    def handle_transaction(self, transaction: Transaction) -> None:
        """
        Processes an incoming transaction in the custom currency.

        First, it fetches the current amount of coins stored on the gateway managed address which should
        already include what has been transferred.
        The amount of coins of the transaction must be determined.

        The further processing depends on whether this is a account based system.
        If not, an internal transaction is done, transferring the received amount to the
        main gateway address.
        Otherwise, the internal transaction is skipped.

        Finally, the waves transaction is performed.
        """
        receivers = self._filter_receivers(transaction)

        if len(receivers) > 1:
            raise MultipleGatewayReceiversError(transaction.tx)

        for i in receivers:
            self._handle_transaction_receiver(transaction.tx, transaction.receivers[i], i, transaction.senders)



"""
MongoKeyValueStorageImpl
"""
from waves_gateway.common import Injectable, KEY_VALUE_STORAGE_COLLECTION
from waves_gateway.model import PollingState
from waves_gateway.serializer import PollingStateSerializer
from waves_gateway.storage.key_value_storage import KeyValueStorage
from pymongo.collection import Collection  # type: ignore
from typing import Optional, Any
from doc_inherit import method_doc_inherit  # type: ignore


@Injectable(deps=[KEY_VALUE_STORAGE_COLLECTION, PollingStateSerializer], provides=KeyValueStorage)
class MongoKeyValueStorageImpl(KeyValueStorage):
    """
    Implements a key value storage with a MongoDB collection.
    """

    _COIN_BLOCK_HEIGHT_KEY = 'coin_block_height'
    _WAVES_BLOCK_HEIGHT_KEY = 'waves_block_height'
    _VALUE_PROPERTY_KEY = 'value'
    _KEY_PROPERTY_KEY = 'key'
    _COIN_POLLING_STATE_KEY = 'coin_polling_state'
    _WAVES_POLLING_STATE_KEY = 'waves_polling_state'

    def _set_value(self, key: str, value: Any) -> None:
        """
        Inserts the key/value pair. Overwrites existing entries.
        """
        query = dict()
        query[MongoKeyValueStorageImpl._KEY_PROPERTY_KEY] = key

        replacement = dict()
        replacement[MongoKeyValueStorageImpl._KEY_PROPERTY_KEY] = key
        replacement[MongoKeyValueStorageImpl._VALUE_PROPERTY_KEY] = value

        self._collection.replace_one(filter=query, replacement=replacement, upsert=True)

    def _get_value(self, key: str) -> Any:
        """
        Returns the value or None if no value was found.
        """
        query = dict()
        query[MongoKeyValueStorageImpl._KEY_PROPERTY_KEY] = key

        query_result = self._collection.find_one(filter=query)

        if query_result is None:
            return None
        else:
            return query_result[MongoKeyValueStorageImpl._VALUE_PROPERTY_KEY]

    @method_doc_inherit
    def set_last_checked_waves_block_height(self, block_height: int) -> None:
        self._set_value(MongoKeyValueStorageImpl._WAVES_BLOCK_HEIGHT_KEY, block_height)

    @method_doc_inherit
    def get_last_checked_waves_block_height(self) -> Optional[int]:
        return self._get_value(MongoKeyValueStorageImpl._WAVES_BLOCK_HEIGHT_KEY)

    def __init__(self, collection: Collection, polling_state_serializer: PollingStateSerializer) -> None:
        self._collection = collection
        self._polling_state_serializer = polling_state_serializer

    @method_doc_inherit
    def set_last_checked_coin_block_height(self, block_height: int) -> None:
        self._set_value(MongoKeyValueStorageImpl._COIN_BLOCK_HEIGHT_KEY, block_height)

    @method_doc_inherit
    def get_last_checked_coin_block_height(self) -> Optional[int]:
        return self._get_value(MongoKeyValueStorageImpl._COIN_BLOCK_HEIGHT_KEY)

    def set_waves_polling_state(self, polling_state: PollingState) -> None:
        self._set_value(MongoKeyValueStorageImpl._WAVES_POLLING_STATE_KEY,
                        self._polling_state_serializer.as_dict(polling_state))

    def get_coin_polling_state(self) -> Optional[PollingState]:
        data = self._get_value(MongoKeyValueStorageImpl._COIN_POLLING_STATE_KEY)

        if data is None:
            return None
        else:
            return self._polling_state_serializer.from_dict(data)

    def get_waves_polling_state(self) -> Optional[PollingState]:
        data = self._get_value(MongoKeyValueStorageImpl._WAVES_POLLING_STATE_KEY)

        if data is None:
            return None
        else:
            return self._polling_state_serializer.from_dict(data)

    def set_coin_polling_state(self, polling_state: PollingState) -> None:
        self._set_value(MongoKeyValueStorageImpl._COIN_POLLING_STATE_KEY,
                        self._polling_state_serializer.as_dict(polling_state))



"""
WavesAddressValidationService
"""

from doc_inherit import class_doc_inherit

from waves_gateway.common import Injectable
from .address_validation_service import AddressValidationService

import pywaves


@Injectable()
@class_doc_inherit
class WavesAddressValidationService(AddressValidationService):
    """inherit"""

    def validate_address(self, address: str) -> bool:
        try:
            pywaves.Address(address=address)
            return True
        except ValueError:
            return False



"""
AttemptListQuery
"""

from typing import Optional


class AttemptListQuery(object):
    """
    An AttemptListQuery summarizes parameters that can be used to search for an TransactionAttemptList.
    """

    def __init__(self,
                 trigger_tx: Optional[str] = None,
                 trigger_currency: Optional[str] = None,
                 trigger_receiver: Optional[int] = None,
                 anything: Optional[str] = None) -> None:
        self._trigger_tx = trigger_tx
        self._trigger_currency = trigger_currency
        self._trigger_receiver = trigger_receiver
        self._anything = anything

    @property
    def trigger_tx(self) -> str:
        return self._trigger_tx

    @property
    def trigger_currency(self) -> str:
        return self._trigger_currency

    @property
    def trigger_receiver(self) -> int:
        return self._trigger_receiver

    @property
    def anything(self) -> str:
        """
        This is a string, the user has entered which could be possibly anything.
        Try to find the best match.
        """
        return self._anything

    def __eq__(self, other):
        res = True

        res = res and self.anything == other.anything
        res = res and self.trigger_tx == other.trigger_tx
        res = res and self.trigger_currency == other.trigger_currency
        res = res and self.trigger_receiver == other.trigger_receiver

        return res



"""
PendingAttemptListSelectorService
"""

from typing import Optional, List

from waves_gateway.common import Injectable
from waves_gateway.storage import TransactionAttemptListStorage
from waves_gateway.model import TransactionAttemptList
import gevent.lock as lock


@Injectable(deps=[TransactionAttemptListStorage])
class PendingAttemptListSelectorService(object):
    """
    This service must be a singleton in the application. It can be used by worker instances
    to acquire an attempt_list. While the attempt_list is being processed by the worker, it cannot be acquired by
    other workers. Once the worker releases the attempt_list, it may be acquired by other workers.
    So, this instance can be called a lock for attempt lists.
    The lock method is secured by a semaphore, so it cannot be called in parallel.
    """

    def __init__(self, attempt_list_storage: TransactionAttemptListStorage) -> None:
        self._attempt_list_storage = attempt_list_storage
        self._lock = lock.Semaphore()
        self._locked_attempt_lists = []  # type: List[str]

    def lock_attempt_list(self) -> Optional[TransactionAttemptList]:
        """
        Locks an attempt list, so that it can be processed.
        """
        self._lock.acquire()

        try:
            result = self._attempt_list_storage.find_oldest_pending_attempt_list()

            if result is not None and result.attempt_list_id not in self._locked_attempt_lists:
                self._locked_attempt_lists.append(result.attempt_list_id)
                result.increment_tries()
                self._attempt_list_storage.update_attempt_list(result)
            else:
                result = None

            self._lock.release()
        finally:
            self._lock.release()

        return result

    def release_attempt_list(self, attempt_list: TransactionAttemptList) -> None:
        """
        Releases an attempt list, so that other workers may work with the attempt list.
        """
        if attempt_list.attempt_list_id in self._locked_attempt_lists:
            self._locked_attempt_lists.remove(attempt_list.attempt_list_id)



"""
List of tokens that are provided by the service module.
"""

from .chain_query_service import ChainQueryService
from waves_gateway.common import InjectionToken
from .address_validation_service import AddressValidationService
from .transaction_service import TransactionService
from .integer_converter_service import IntegerConverterService
from .transaction_polling_service import TransactionPollingService

COIN_ADDRESS_VALIDATION_SERVICE = InjectionToken('COIN_ADDRESS_VALIDATION_SERVICE', AddressValidationService)
COIN_CHAIN_QUERY_SERVICE = InjectionToken('COIN_CHAIN_QUERY_SERVICE', ChainQueryService)
WAVES_CHAIN_QUERY_SERVICE = InjectionToken('WAVES_CHAIN_QUERY_SERVICE', ChainQueryService)
COIN_TRANSACTION_SERVICE = InjectionToken('COIN_TRANSACTION_SERVICE', TransactionService)
COIN_INTEGER_CONVERTER_SERVICE = InjectionToken('COIN_INTEGER_CONVERTER_SERVICE', IntegerConverterService)
ASSET_INTEGER_CONVERTER_SERVICE = InjectionToken('ASSET_INTEGER_CONVERTER_SERVICE', IntegerConverterService)
COIN_TRANSACTION_SERVICE_CONVERTER_PROXY = InjectionToken('COIN_TRANSACTION_SERVICE_CONVERTER_PROXY',
                                                          TransactionService)
WAVES_TRANSACTION_SERVICE_CONVERTER_PROXY = InjectionToken('WAVES_TRANSACTION_SERVICE_CONVERTER_PROXY',
                                                           TransactionService)
COIN_TRANSACTION_POLLING_SERVICE = InjectionToken('COIN_TRANSACTION_POLLING_SERVICE', TransactionPollingService)
WAVES_TRANSACTION_POLLING_SERVICE = InjectionToken('WAVES_TRANSACTION_POLLING_SERVICE', TransactionPollingService)



"""
MongoMapStorageImpl
"""

from typing import Optional

from doc_inherit import method_doc_inherit  # type: ignore
from pymongo.collection import Collection  # type: ignore

from waves_gateway.common import Injectable, MAP_STORAGE_COLLECTION
from waves_gateway.model import MappingEntry
from waves_gateway.serializer import MappingEntrySerializer
from .map_storage import MapStorage


@Injectable(deps=[MAP_STORAGE_COLLECTION, MappingEntrySerializer], provides=MapStorage)
class MongoMapStorageImpl(MapStorage):
    """
    Implementation of a MapStorage using a MongoDB.
    """

    @method_doc_inherit
    def save_mapping(self, mapping: MappingEntry):
        self._collection.insert_one(self._serializer.as_dict(mapping))

    def __init__(self, collection: Collection, serializer: MappingEntrySerializer) -> None:
        super().__init__()
        self._collection = collection
        self._serializer = serializer

    @method_doc_inherit
    def get_waves_address_by_coin_address(self, coin_address: str) -> Optional[str]:
        query = dict()
        query[MappingEntry.DICT_COIN_KEY] = coin_address

        query_result = self._collection.find_one(query)
        if query_result is None:
            return None
        else:
            return str(query_result[MappingEntry.DICT_WAVES_KEY])

    @method_doc_inherit
    def get_coin_address_by_waves_address(self, waves_address: str) -> Optional[str]:
        query = dict()
        query[MappingEntry.DICT_WAVES_KEY] = waves_address
        query_result = self._collection.find_one(query)
        if query_result is None:
            return None
        else:
            return str(query_result[MappingEntry.DICT_COIN_KEY])



"""Serializer"""

from abc import ABC, abstractmethod
from typing import Any


class Serializer(ABC):
    """Base class for all serializers. Ensures a similar API."""

    @abstractmethod
    def as_dict(self, value: Any) -> dict:
        pass

    def from_dict(self, data: dict) -> Any:
        pass



"""
Offers an optional API to the Gateway
"""

from .flask_rest_controller import FlaskRestController
from .gateway_controller import GatewayController
from .gateway_controller_impl import GatewayControllerImpl
from .gateway_controller_converter_proxy_impl import GatewayControllerConverterProxyImpl



from distutils.core import setup
from setuptools import find_packages
from distutils.cmd import Command


class CoverageCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys, subprocess

        code = subprocess.call([
            sys.executable, '-m', 'nose', 'tests', '--with-coverage', '--cover-package', 'waves_gateway',
            '--cover-html'
        ])

        exit(code)


class MyPyCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys, subprocess

        code = subprocess.call([sys.executable, '-m', 'mypy', 'waves_gateway', '--ignore-missing-imports'])

        exit(code)


class LintCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys, subprocess

        code = subprocess.call([sys.executable, '-m', 'pylint', 'waves_gateway'])

        exit(code)


class DocsCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys, subprocess, os, shutil

        if os.path.exists("docs"):
            shutil.rmtree("docs")

        if os.path.exists(".apidoc"):
            shutil.rmtree(".apidoc")

        code = subprocess.call(['sphinx-apidoc', '-o', '.apidoc', 'waves_gateway', '-l', '-F', '-H', 'Waves-Gateway-Framework', '-A', 'Henning Gerrits', '-f'])

        if code != 0:
            exit(code)

        code = subprocess.call(['sphinx-build', '.apidoc', 'docs'])

        open('docs/.nojekyll', 'a').close()

        exit(code)


setup(
    name='waves_gateway',
    url='https://github.com/jansenmarc/WavesGatewayFramework',
    version='1.0.1',
    author='Henning Gerrits',
    test_suite='nose.collector',
    tests_require=['nose'],
    keywords='waves gateway wavesplatform',
    python_requires='>=3.5',
    install_requires=[
        'PyWaves>=0.7.9', 'python-doc-inherit>=0.3.0', 'simplejson>=3.11.1', 'requests>=2.9.1', 'base58>=0.2.5',
        'pymongo>=3.4.0', 'Flask>=0.12.2', 'gevent>=1.2.2'
    ],
    description='A framework to connect other cryptocurrencies to the Waves-Platform.',
    package_data={'waves_gateway': ['static/**/*', 'static/*']},
    license='MIT',
    long_description=open('README.rst').read(),
    packages=find_packages(exclude=['tests']),
    cmdclass={
        'coverage': CoverageCommand,
        'mypy': MyPyCommand,
        'lint': LintCommand,
        'docs': DocsCommand
    })



"""
TransactionPollingService
"""

import logging

import time
import gevent
import gevent.pool as pool
import gevent.greenlet as greenlet
from typing import List, Optional

from waves_gateway.common import filter_array
from waves_gateway.model import Transaction, PollingDelayConfig, PollingState, PollingTransactionState
from .transaction_consumer import TransactionConsumer
from waves_gateway.storage import BlockHeightStorageProxy, PollingStateStorageProxy

from .chain_query_service import ChainQueryService


class TransactionPollingService(greenlet.Greenlet):
    """
    The polling service is intended to be spawned once on the start of the application.
    It will reschedule executions of itself in dynamic intervals. It is possible, to cancel
    this service. In this case, no further executions will be scheduled.
    The delay between the executions tries to adjust itself to the time between the creation of new blocks.
    But it will never be less than the min_polling_delay nor more than the max_polling_delay.
    

    It is designed, so that it can be used for both Waves and the custom currency.
    For the particular case, the service must be constructed with the matching parameters
    for the currency.
    """

    def __init__(self, chain_query_service: ChainQueryService, logger: logging.Logger,
                 block_height_storage: BlockHeightStorageProxy, transaction_consumer: TransactionConsumer,
                 polling_state_storage: PollingStateStorageProxy, max_handle_transaction_tries: int,
                 min_polling_delay_s: float, max_polling_delay_s: float, last_block_distance: int) -> None:
        greenlet.Greenlet.__init__(self)

        self._consumer = transaction_consumer
        self._chain_query_service = chain_query_service
        self._logger = logger.getChild(self.__class__.__name__ + "." + transaction_consumer.__class__.__name__)
        self._block_height_storage = block_height_storage
        self._cancelled = False
        self._polling_delay_s = min_polling_delay_s
        self._d_height = 0
        self._min_polling_delay = min_polling_delay_s
        self._max_polling_delay = max_polling_delay_s
        self._last_block_time_s = None  # type: Optional[float]
        self._last_polling_delay = self._polling_delay_s
        self._polling_state_storage = polling_state_storage
        self._max_handle_transaction_tries = max_handle_transaction_tries
        self._polling_state = PollingState()
        self._current_height = None
        self._last_block_distance = last_block_distance

    def _transaction_not_exceeds_tries(self, transaction: Transaction) -> bool:
        """Returns whether the given transaction has failed too often."""
        return self._polling_state.transaction_map[transaction.tx].tries < self._max_handle_transaction_tries

    def _is_transaction_not_already_processed(self, transaction: Transaction) -> bool:
        """Returns whether the transaction has already been successfully processed."""
        transaction_state = self._polling_state.transaction_map[transaction.tx]
        return not transaction_state.ok

    def _should_transaction_be_processed(self, tr: Transaction) -> bool:
        """Summarizes every necessary condition for a transaction to be processed."""
        return self._consumer.filter_transaction(tr) and self._is_transaction_not_already_processed(
            tr) and self._transaction_not_exceeds_tries(tr)

    def _filter_transaction_task(self, transaction: Transaction) -> Optional[Transaction]:
        """Returns the transaction if it should be processed or None otherwise."""
        self._ensure_polling_state_has_transaction(transaction)

        if self._should_transaction_be_processed(transaction):
            return transaction
        else:
            return None

    def _filter_transactions(self, transactions: List[Transaction]) -> List[Transaction]:
        """Filters all transactions in parallel. The results may not have the same order."""
        filter_task_group = pool.Group()
        filter_task_results = filter_task_group.imap_unordered(self._filter_transaction_task, transactions)
        return filter_array(lambda el: el is not None, filter_task_results)

    def cancel(self):
        """Marks the polling as cancelled."""
        self._cancelled = True

    def _log_delay(self):
        self._logger.debug('Sleeping %ss', str(self._polling_delay_s))

    def _apply_delay_border(self):
        if self._polling_delay_s < self._min_polling_delay:
            self._polling_delay_s = self._min_polling_delay

        if self._polling_delay_s > self._max_polling_delay:
            self._polling_delay_s = self._max_polling_delay

    def _adjust_delay_after_not_found_block(self):
        self._polling_delay_s = round(self._polling_delay_s + (self._max_polling_delay * 0.01), 2)
        self._apply_delay_border()

    def _adjust_delay_after_found_block(self):
        if self._last_block_time_s is None:
            self._polling_delay_s = self._min_polling_delay
        elif self._d_height > 1:
            self._polling_delay_s = round(self._polling_delay_s / 2, 2)
        else:
            current_time_s = time.time()
            diff = current_time_s - self._last_block_time_s

            if (self._polling_delay_s + diff) > 0:
                self._polling_delay_s = round((self._polling_delay_s + diff) / 6, 2)

        self._last_block_time_s = time.time()
        self._apply_delay_border()

    def _adjust_delay_after_exception(self):
        self._polling_delay_s = self._max_polling_delay

    def _run(self):
        try:
            while not self._cancelled:
                try:
                    self._iteration()
                except BaseException as ex:
                    self._logger.error(ex, exc_info=True)
                    self._logger.debug("Execution has finished with errors")
                    self._adjust_delay_after_exception()
                finally:
                    self._log_delay()

                gevent.sleep(self._polling_delay_s)
        finally:
            self._logger.info("Cancelled")

    def _ensure_polling_state_has_transaction(self, tr: Transaction):
        if tr.tx not in self._polling_state.transaction_map:
            self._polling_state.transaction_map[tr.tx] = PollingTransactionState()

    def _handle_transaction(self, transaction: Transaction):
        """Calls the transaction_consumer with the given transaction."""
        self._logger.debug("Overhand transaction %s to consumer %s.", transaction.tx, self._consumer.__class__.__name__)

        self._ensure_polling_state_has_transaction(transaction)

        try:
            self._consumer.handle_transaction(transaction)
            self._polling_state.transaction_map[transaction.tx].mark_as_done()
        except BaseException as ex:
            self._polling_state.transaction_map[transaction.tx].increment_tries()
            raise ex

    def _fetch_stored_polling_state(self):
        self._polling_state = self._polling_state_storage.get_polling_state()

        if self._polling_state is None:
            self._polling_state = PollingState()

    def _reset_polling_state(self):
        self._polling_state = PollingState()

    def _update_stored_polling_state(self):
        self._polling_state_storage.set_polling_state(self._polling_state)

    def _handle_transactions(self, transactions: List[Transaction]) -> None:
        handle_task_group = pool.Group()

        for tr in transactions:
            handle_task_group.spawn(lambda: self._handle_transaction(tr))

        handle_task_group.join(raise_error=True)

    def _iteration(self):
        """
        In every execution, it first fetches the last checked block height from storage.
        Along that, it needs the overall height of the chain.
        If the storage is empty, the service only fetches the latest block of the chain.
        If not, which means that there is a last checked block height defined, the service checks all
        blocks from this height to the maximum height.
        When finished with all these blocks, the current maximum height is stored as the last
        checked block height in storage.

        For every block, it fetches all transactions. Those transactions are filtered using the filter
        functionality of the transaction consumer instance.
        Finally, the filtered transactions are each overhanded to the consumer.
        """

        height_of_last_checked_block = self._block_height_storage.get_last_checked_block_height()
        height_of_highest_block = self._chain_query_service.get_height_of_highest_block()

        self._fetch_stored_polling_state()

        if height_of_last_checked_block is None:
            height_of_last_checked_block = height_of_highest_block
            self._block_height_storage.set_last_checked_block_height(height_of_last_checked_block)

        self._d_height = height_of_highest_block - height_of_last_checked_block

        if height_of_last_checked_block >= height_of_highest_block - self._last_block_distance:
            self._logger.debug("Skipping run of transaction polling service as no new blocks are available.")
            self._adjust_delay_after_not_found_block()
            return
        else:
            self._adjust_delay_after_found_block()

        current_height = height_of_last_checked_block + 1

        self._logger.info("Checking block at height %s", str(current_height))

        transactions = self._chain_query_service.get_transactions_of_block_at_height(current_height)
        filtered_transactions = self._filter_transactions(transactions)

        try:
            self._handle_transactions(filtered_transactions)
            self._reset_polling_state()
        finally:
            self._update_stored_polling_state()

        self._block_height_storage.set_last_checked_block_height(current_height)

        self._logger.debug("Finished block at height %s", str(current_height))



"""
TransactionAttemptListStorage
"""

from abc import ABC, abstractmethod
from typing import Optional, List

import gevent.lock

from waves_gateway.common import DuplicateAttemptListTriggerError, DuplicateAttemptListIDError
from waves_gateway.model import TransactionAttemptList, AttemptListTrigger, AttemptListQuery


class TransactionAttemptListStorage(ABC):
    """
    Defines what an storage for TransactionAttemptList instances must be capable of.
    """

    def __init__(self):
        self._save_lock = gevent.lock.Semaphore()

    @abstractmethod
    def find_by_trigger(self, trigger: AttemptListTrigger) -> Optional[TransactionAttemptList]:
        """
        Finds a specific attempt_list by its trigger.
        """
        pass

    @abstractmethod
    def find_by_attempt_list_id(self, attempt_list_id: str) -> Optional[TransactionAttemptList]:
        """
        Finds a specific attempt_list by its trigger.
        """
        pass

    @abstractmethod
    def trigger_exists(self, trigger: AttemptListTrigger) -> bool:
        """
        Tests if an attempt_list with the given trigger exists.
        """
        pass

    @abstractmethod
    def attempt_list_id_exists(self, attempt_list_id: str) -> bool:
        """
        Tests if an attempt_list with the given id exists.
        """
        pass

    @abstractmethod
    def gateway_transaction_exists(self, tx: str) -> bool:
        """
        Tests if a given transaction identifier exists.
        """
        pass

    @abstractmethod
    def save_attempt_list(self, attempt_list: TransactionAttemptList) -> None:
        """
        Creates a new attempt_list.
        """
        pass

    def safely_save_attempt_list(self, attempt_list: TransactionAttemptList) -> None:
        """
        Creates a new attempt_list. Throws an exception if the id or the trigger already exist.
        """
        self._save_lock.acquire()

        try:
            if self.trigger_exists(attempt_list.trigger):
                raise DuplicateAttemptListTriggerError()

            if self.attempt_list_id_exists(attempt_list.attempt_list_id):
                raise DuplicateAttemptListIDError()

            self.save_attempt_list(attempt_list)
            self._save_lock.release()
        finally:
            self._save_lock.release()

    @abstractmethod
    def update_attempt_list(self, attempt_list: TransactionAttemptList) -> None:
        """
        Updates an existing attempt_list.
        """
        pass

    @abstractmethod
    def query_attempt_lists(self, query: AttemptListQuery) -> List[TransactionAttemptList]:
        """
        Searches for attempt lists by the given criteria.
        """
        pass

    @abstractmethod
    def find_oldest_pending_attempt_list(self) -> Optional[TransactionAttemptList]:
        """
        Queries for the oldest attempt list that is not complete.
        """
        pass



"""
Defines serializers for the data models defined in the waves_gateway.model module.
Those serializers may be used to persist the models.
"""

from .key_pair_serializer import KeyPairSerializer
from .mapping_entry_serializer import MappingEntrySerializer
from .transaction_attempt_list_serializer import TransactionAttemptListSerializer
from .public_configuration_serializer import PublicConfigurationSerializer
from .polling_state_serializer import PollingStateSerializer
from .polling_transaction_state_serializer import PollingTransactionStateSerializer
from .serializer import Serializer



"""GatewayConfigFile"""

from typing import Optional, Union
from decimal import Decimal
from .key_pair import KeyPair


class GatewayConfigFile(object):
    """
    Contains the information of a parsed configuration file.
    """

    def __init__(self):
        self.waves_node = None  # type: Optional[str]
        self.coin_node = None  # type: Optional[str]
        self.coin_fee = None  # type: Optional[Union[int, Decimal]]
        self.gateway_fee = None  # type: Optional[Union[int, Decimal]]
        self.gateway_owner_address = None  # type: Optional[str]
        self.gateway_waves_address_secret = None  # type: Optional[KeyPair]
        self.gateway_coin_address_secret = None  # type: Optional[KeyPair]
        self.mongo_host = None  # type: Optional[str]
        self.mongo_port = None  # type: Optional[int]
        self.mongo_database = None  # type: Optional[str]
        self.waves_chain = None  # type: Optional[str]
        self.environment = None  # type: Optional[str]
        self.gateway_host = None  # type: Optional[str]
        self.gateway_port = None  # type: Optional[int]



"""
GatewayControllerImpl
"""

from doc_inherit import method_doc_inherit, class_doc_inherit  # type: ignore

from logging import Logger
from typing import Optional, List

from waves_gateway.common import WavesAddressInvalidError, InvalidTransactionIdentifier, Injectable, \
    COIN_CHAIN_QUERY_SERVICE_CONVERTER_PROXY, WAVES_CHAIN_QUERY_SERVICE_CONVERTER_PROXY
from waves_gateway.factory.coin_address_factory import CoinAddressFactory
from waves_gateway.model import MappingEntry, KeyPair, AttemptListTrigger, \
    TransactionAttemptList, AttemptListQuery
from waves_gateway.service import ChainQueryService, TransactionConsumer, \
    AddressValidationService, WavesAddressValidationService, WavesTransactionConsumerImpl, \
    CoinTransactionConsumerImpl
from waves_gateway.storage import TransactionAttemptListStorage
from waves_gateway.storage.map_storage import MapStorage
from waves_gateway.storage.wallet_storage import WalletStorage
from .gateway_controller import GatewayController


@Injectable(
    provides=GatewayController,
    deps=[
        CoinAddressFactory, Logger, MapStorage, WalletStorage, TransactionAttemptListStorage,
        WavesAddressValidationService, WAVES_CHAIN_QUERY_SERVICE_CONVERTER_PROXY,
        COIN_CHAIN_QUERY_SERVICE_CONVERTER_PROXY, CoinTransactionConsumerImpl, WavesTransactionConsumerImpl
    ])
@class_doc_inherit
class GatewayControllerImpl(GatewayController):
    """inherit"""

    def __init__(self, coin_address_factory: CoinAddressFactory, logger: Logger, map_storage: MapStorage,
                 wallet_storage: WalletStorage, attempt_list_storage: TransactionAttemptListStorage,
                 waves_address_validation_service: AddressValidationService,
                 waves_chain_query_service: ChainQueryService, coin_chain_query_service: ChainQueryService,
                 coin_transaction_consumer: TransactionConsumer,
                 waves_transaction_consumer: TransactionConsumer) -> None:
        self._coin_address_factory = coin_address_factory
        self._map_storage = map_storage
        self._wallet_storage = wallet_storage
        self._logger = logger.getChild('GatewayControllerImpl')
        self._attempt_list_storage = attempt_list_storage
        self._waves_address_validation_service = waves_address_validation_service
        self._waves_chain_query_service = waves_chain_query_service
        self._coin_chain_query_service = coin_chain_query_service
        self._coin_transaction_consumer = coin_transaction_consumer
        self._waves_transaction_consumer = waves_transaction_consumer

    @method_doc_inherit
    def check_waves_transaction(self, tx: str) -> None:
        transaction = self._waves_chain_query_service.get_transaction_by_tx(tx)

        if transaction is None:
            raise InvalidTransactionIdentifier()

        if self._waves_transaction_consumer.filter_transaction(transaction):
            self._waves_transaction_consumer.handle_transaction(transaction)

    @method_doc_inherit
    def check_coin_transaction(self, tx: str) -> None:
        transaction = self._coin_chain_query_service.get_transaction_by_tx(tx)

        if transaction is None:
            raise InvalidTransactionIdentifier()

        if self._coin_transaction_consumer.filter_transaction(transaction):
            self._coin_transaction_consumer.handle_transaction(transaction)

    @method_doc_inherit
    def validate_waves_address(self, address: str) -> bool:
        """inherit"""
        return self._waves_address_validation_service.validate_address(address)

    @method_doc_inherit
    def create_address(self, waves_address: str) -> str:
        """inherit"""
        self._logger.debug("Client requested coin address for waves address '%s'", waves_address)

        if not self._waves_address_validation_service.validate_address(waves_address):
            raise WavesAddressInvalidError()

        if self._map_storage.waves_address_exists(waves_address):
            coin_address = self._map_storage.get_coin_address_by_waves_address(waves_address)
            self._logger.debug("Found associated coin_address '%s'", coin_address)
            return coin_address
        else:
            create_address_result = self._coin_address_factory.create_address()

            if isinstance(create_address_result, str):
                mapping = MappingEntry(waves_address, create_address_result)
                self._map_storage.safely_save_mapping(mapping)
                self._logger.info("Created new mapping '%s'", str(mapping))
                return create_address_result
            elif isinstance(create_address_result, KeyPair):
                mapping = MappingEntry(waves_address, create_address_result.public)
                self._wallet_storage.safely_save_address_secret(create_address_result)
                self._map_storage.safely_save_mapping(mapping)
                self._logger.info("Created new mapping '%s'", str(mapping))

                return create_address_result.public
            else:
                raise TypeError('Result of create_address is neither a CoinAddress nor a CoinAddressSecret')

    @method_doc_inherit
    def get_attempt_list_by_id(self, attempt_list_id: str) -> Optional[TransactionAttemptList]:
        return self._attempt_list_storage.find_by_attempt_list_id(attempt_list_id)

    @method_doc_inherit
    def query_attempt_lists(self, query: AttemptListQuery) -> List[TransactionAttemptList]:
        return self._attempt_list_storage.query_attempt_lists(query)

    @method_doc_inherit
    def get_attempt_list_by_trigger(self, trigger: AttemptListTrigger) -> Optional[TransactionAttemptList]:
        return self._attempt_list_storage.find_by_trigger(trigger)
